import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype

import matplotlib.pyplot as plt
from matplotlib import rcParams
import seaborn as sns

from .utils import *
from .utils import _make_iterable, _is_iter


def _plot_grid(plot_func, data, x, y, color=None, **kwargs):
    MSG_ONLY_2D = "Only 2 of 3 dimensions can be iterables to create grids: x axis, y axis, or color."

    # Determine split_axis
    if _is_iter(x):
        if _is_iter(y):
            assert not _is_iter(color), MSG_ONLY_2D
        elif _is_iter(color):
            assert not _is_iter(y), MSG_ONLY_2D
        else:
            return _plot_grid_from_1d(plot_func, data, x, y, color, "x", **kwargs)
        return _plot_2d_grid(plot_func, data, x, y, color, **kwargs)
    elif _is_iter(y):
        if _is_iter(x):
            assert not _is_iter(color), MSG_ONLY_2D
        elif _is_iter(color):
            assert not _is_iter(x), MSG_ONLY_2D
        else:
            return _plot_grid_from_1d(plot_func, data, x, y, color, "y", **kwargs)
        return _plot_2d_grid(plot_func, data, x, y, color, **kwargs)
    elif _is_iter(color):
        if _is_iter(x):
            assert not _is_iter(y), MSG_ONLY_2D
        elif _is_iter(y):
            assert not _is_iter(x), MSG_ONLY_2D
        else:
            return _plot_grid_from_1d(plot_func, data, x, y, color, "color", **kwargs)
        return _plot_2d_grid(plot_func, data, x, y, color, **kwargs)
    else:
        return _plot_grid_from_1d(plot_func, data, x, y, [color], "color", **kwargs)


def _plot_grid_from_1d(
    plot_func,
    data,
    x,
    y,
    color=None,
    split_axis="color",  # x, y, or color
    ncols=4,
    zero_line_x=False,
    zero_line_y=False,
    linewidth=0,
    zero_linewidth=1,
    legend=True,
    legend_prop=None,
    palette=None,
    sharex=False,
    sharey=False,
    modifier=None,
    **kwargs,
):
    x = maybe_factor_indices_to_factors(x)
    y = maybe_factor_indices_to_factors(y)
    color = maybe_factor_indices_to_factors(color)

    # Define the variable for the split
    # and plot axes labels
    if split_axis == "color":
        split_vars = maybe_factor_indices_to_factors(_make_iterable(color))
        x_label = x
        y_label = y

    elif split_axis == "x":
        split_vars = maybe_factor_indices_to_factors(_make_iterable(x))
        y_label = y

    elif split_axis == "y":
        split_vars = maybe_factor_indices_to_factors(_make_iterable(y))
        x_label = x

    # Set default colour to black if none set
    if "c" not in kwargs and (color is None or color == [None]):
        kwargs["color"] = "black"

    legend_str = "brief" if (legend and color is not None) else False

    # Figure out rows & columns for the grid with plots
    ncols = min(ncols, len(split_vars))
    nrows = int(np.ceil(len(split_vars) / ncols))
    fig, axes = plt.subplots(
        nrows,
        ncols,
        sharex=sharex,
        sharey=sharey,
        figsize=(
            ncols * rcParams["figure.figsize"][0],
            nrows * rcParams["figure.figsize"][1],
        ),
    )
    if ncols == 1:
        axes = np.array(axes).reshape(-1, 1)
    if nrows == 1:
        axes = np.array(axes).reshape(1, -1)

    for i, split_var in enumerate(split_vars):
        color_var = color if split_axis != "color" else split_var

        ri = i // ncols
        ci = i % ncols

        # data_ = data.sort_values(color) if color is not None and color != [None] else data

        with sns.axes_style("ticks"), sns.color_palette(palette or "Set2"):

            g = plot_func(
                x=x if split_axis != "x" else split_var,
                y=y if split_axis != "y" else split_var,
                data=data,
                hue=color_var,
                linewidth=linewidth,
                legend=legend_str,
                palette=palette,
                ax=axes[ri, ci],
                **kwargs,
            )
            # Otherwise sns.violinplot still plots the legend
            if legend_str is False:
                try:
                    g.get_legend().remove()
                except AttributeError:
                    pass

            if modifier:
                modifier(split_var=split_var, color_var=color_var, ax=g)

            sns.despine(offset=10, trim=True, ax=g)

            if split_axis == "x":
                x_label = f"Factor{split_var+1}" if isinstance(x, int) else split_var
            elif split_axis == "y":
                y_label = f"Factor{split_var+1}" if isinstance(y, int) else split_var
            g.set(
                xlabel=f"{x_label}",
                ylabel=f"{y_label}",
                title=split_var,
            )

            if legend and color_var:
                if is_numeric_dtype(data[color_var]):
                    means = data.groupby(color_var)[color_var].mean()
                    norm = plt.Normalize(means.min(), means.max())
                    cmap = (
                        palette
                        if palette is not None
                        else sns.cubehelix_palette(as_cmap=True)
                    )
                    sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
                    sm.set_array([])
                    try:
                        g.figure.colorbar(sm, ax=axes[ri, ci])
                        g.get_legend().remove()
                    except Exception:
                        warn("Cannot make a proper colorbar")
                else:
                    g.legend(
                        bbox_to_anchor=(1.05, 1),
                        loc=2,
                        borderaxespad=0.0,
                        prop=legend_prop,
                    )

            if zero_line_y:
                axes[ri, ci].axhline(
                    0, ls="--", color="lightgrey", linewidth=zero_linewidth, zorder=0
                )
            if zero_line_x:
                axes[ri, ci].axvline(
                    0, ls="--", color="lightgrey", linewidth=zero_linewidth, zorder=0
                )

        # Remove unused axes
        for i in range(len(split_vars), ncols * nrows):
            ri = i // ncols
            ci = i % ncols
            fig.delaxes(axes[ri, ci])

        plt.tight_layout()

    return g


def _plot_2d_grid(*args, **kwargs):
    raise NotImplementedError



from .core import mofa_model
from .utils import *

import sys
from warnings import warn
from typing import Union, Optional, List, Iterable, Sequence
from functools import partial

import numpy as np
from scipy.stats import pearsonr
import pandas as pd
from pandas.api.types import is_numeric_dtype
import matplotlib.pyplot as plt
from matplotlib import rcParams
import seaborn as sns

from .utils import maybe_factor_indices_to_factors, _make_iterable, _is_iter
from .plot_utils import _plot_grid


### FACTORS ###


def plot_factors_scatter(
    model: mofa_model,
    x="Factor1",
    y="Factor2",
    dist=False,
    groups=None,
    group_label="group",
    color=None,
    zero_line_x=False,
    zero_line_y=False,
    linewidth=0,
    zero_linewidth=1,
    size=20,
    legend=True,
    legend_prop=None,
    palette=None,
    ncols=4,
    sharex=False,
    sharey=False,
    **kwargs,
):
    """
    Plot samples features such as factor values,
    samples metadata or covariates

    Parameters
    ----------
    model : mofa_model
        Factor model
    x : optional
        Factor or variable (metadata column, feature, or covariate) to plot along X axis (Factor1 by default).
        A list can be provided to create a grid of plots.
    y : optional
        Factor or variable (metadata column, feature, or covariate) to plot along Y axis (Factor2 by default).
        A list can be provided to create a grid of plots.
    dist : optional
        Boolean value if to add marginal distributions or histograms to the scatterplot (jointplot)
    groups : optional
        Subset of groups to consider
    group_label : optional
        Sample (cell) metadata column to be used as group assignment ('group' by default)
    color : optional
        Grouping variable by default, alternatively a feature name can be provided (when no kde).
        If a list of features is provided, they will be plot on one figure.
        Use palette argument to provide a colour map.
    zero_line : optional
        Boolean values if to add Z=0 line
    linewidth : optional
        Linewidth argument for dots (default is 0)
    zero_linewidth : optional
        Linewidth argument for the zero line (default is 1)
    size : optional
        Size argument for dots (ms for plot, s for jointplot and scatterplot; default is 5)
    legend : optional bool
        If to show the legend (e.g. colours matching groups)
    legend_prop : optional
        The font properties of the legend
    palette : optional
        cmap describing colours, default is None (cubehelix)
        Example palette: seaborn.cubehelix_palette(8, start=.5, rot=-.75. as_cmap=True)
    ncols : optional
        Number of columns if multiple colours are defined (4 by default)
    sharex: optional
        Common X axis across plots on the grid
    sharey: optional
        Common Y axis across plots on the grid
    """

    # Process input arguments
    if group_label == "group" and color is None:
        color = "group"
    color_vars = maybe_factor_indices_to_factors(_make_iterable(color))

    assert not (
        len(color_vars) > 1 and dist
    ), "When plotting distributions, only one color can be provided"

    assert not (
        (_is_iter(x) or _is_iter(y)) and dist
    ), "When plotting distributions, only scalar x and y axes can be defined"

    if dist:
        # Get values
        z = model.fetch_values([x, y], unique=True)

        # Add group and colour information
        vars = [group_label, *color_vars]
        vars = [v for v in vars if v not in z.columns]
        if any([not (not (i)) for i in vars]):
            meta = model.fetch_values(variables=vars)
            z = z.rename_axis("sample").reset_index()
            z = z.set_index("sample").join(meta).reset_index()

        # Subset groups (incl. custom groups of samples)
        if group_label and groups is not None:
            z = z[z[group_label].isin(groups)]

        g = sns.jointplot(
            x=x,
            y=y,
            hue=color_vars[0],
            data=z.sort_values(color_vars[0]) if color_vars[0] else z,
            linewidth=linewidth,
            s=size,
            legend=legend_str,
            palette=palette,
            **kwargs,
        )
        sns.despine(offset=10, trim=True, ax=g.ax_joint)
        g.ax_joint.set(
            xlabel=f"{x_factor_label} value", ylabel=f"{y_factor_label} value"
        )
        if legend:
            g.ax_joint.legend(
                bbox_to_anchor=(1.4, 1), loc=2, borderaxespad=0.0, prop=legend_prop
            )
        if zero_line_y:
            g.axhline(0, ls="--", color="lightgrey", linewidth=linewidth, zorder=0)
        if zero_line_x:
            g.axvline(0, ls="--", color="lightgrey", linewidth=linewidth, zorder=0)
        plt.tight_layout()
    else:
        plot = partial(
            sns.scatterplot,
            s=size,
        )
        g = _plot_factors(
            plot,
            model,
            x,
            y,
            color,
            groups=groups,
            group_label=group_label,
            zero_line_x=zero_line_x,
            zero_line_y=zero_line_y,
            linewidth=linewidth,
            zero_linewidth=zero_linewidth,
            legend=legend,
            legend_prop=legend_prop,
            palette=palette,
            ncols=ncols,
            sharex=sharex,
            sharey=sharey,
            **kwargs,
        )

    return g


plot_factors = plot_factors_scatter


def _plot_factors(
    plot_func,
    model: mofa_model,
    x="Factor1",
    y="Factor2",
    color=None,
    groups=None,
    group_label="group",
    zero_line_x=False,
    zero_line_y=False,
    linewidth=0,
    zero_linewidth=1,
    legend=True,
    legend_prop=None,
    palette=None,
    ncols=4,
    sharex=False,
    sharey=False,
    **kwargs,
):
    """
    Plot samples features such as factor values,
    samples metadata or covariates

    Parameters
    ----------
    model : mofa_model
        Factor model
    x : optional
        Factor or variable (metadata column, feature, or covariate) to plot along X axis (Factor1 by default).
        A list can be provided to create a grid of plots.
    y : optional
        Factor or variable (metadata column, feature, or covariate) to plot along Y axis (Factor2 by default).
        A list can be provided to create a grid of plots.
    dist : optional
        Boolean value if to add marginal distributions or histograms to the scatterplot (jointplot)
    groups : optional
        Subset of groups to consider
    group_label : optional
        Sample (cell) metadata column to be used as group assignment ('group' by default)
    color : optional
        Grouping variable by default, alternatively a feature name can be provided (when no kde).
        If a list of features is provided, they will be plot on one figure.
        Use palette argument to provide a colour map.
    zero_line : optional
        Boolean values if to add Z=0 line
    linewidth : optional
        Linewidth argument for dots (default is 0)
    zero_linewidth : optional
        Linewidth argument for the zero line (default is 1)
    size : optional
        Size argument for dots (ms for plot, s for jointplot and scatterplot; default is 5)
    legend : optional bool
        If to show the legend (e.g. colours matching groups)
    legend_prop : optional
        The font properties of the legend
    palette : optional
        cmap describing colours, default is None (cubehelix)
        Example palette: seaborn.cubehelix_palette(8, start=.5, rot=-.75. as_cmap=True)
    ncols : optional
        Number of columns if multiple colours are defined (4 by default)
    sharex: optional
        Common X axis across plots on the grid
    sharey: optional
        Common Y axis across plots on the grid
    """

    # Process input arguments
    if group_label == "group" and color is None:
        color = "group"
    color_vars = maybe_factor_indices_to_factors(_make_iterable(color))

    # Accept x and y as lists to create a grid
    x_vars = maybe_factor_indices_to_factors(_make_iterable(x))
    y_vars = maybe_factor_indices_to_factors(_make_iterable(y))

    # Get values
    z = model.fetch_values([*x_vars, *y_vars], unique=True)

    # Add group and colour information
    vars = [group_label, *color_vars]
    vars = [v for v in vars if v not in z.columns]
    if any([not (not (i)) for i in vars]):
        meta = model.fetch_values(variables=vars)
        z = z.rename_axis("sample").reset_index()
        z = z.set_index("sample").join(meta).reset_index()

    # Subset groups (incl. custom groups of samples)
    if group_label and groups is not None:
        z = z[z[group_label].isin(groups)]

    g = _plot_grid(
        plot_func,
        z,
        x,
        y,
        color,
        zero_line_x=zero_line_x,
        zero_line_y=zero_line_y,
        linewidth=linewidth,
        zero_linewidth=zero_linewidth,
        legend=legend,
        legend_prop=legend_prop,
        palette=palette,
        ncols=ncols,
        sharex=sharex,
        sharey=sharey,
        **kwargs,
    )

    return g


def plot_factors_violin(
    model: mofa_model,
    factors: Union[int, List[int]] = None,
    color="group",
    violins=True,
    dots=False,
    zero_line=True,
    group_label="group",
    groups=None,
    linewidth=0,
    zero_linewidth=1,
    size=20,
    legend=True,
    legend_prop=None,
    palette=None,
    alpha=None,
    violins_alpha=None,
    ncols=4,
    sharex=False,
    sharey=False,
    **kwargs,
):
    """
    Plot factor values as violinplots or stripplots (jitter plots)

    Parameters
    ----------
    model : mofa_model
        Factor model
    factors : optional
        Index of a factor (or indices of factors) to use (all factors by default)
    x : optional
        Variable to plot along X axis (factor identity by default)
    y : optional
        Variable to plot along Y axis (factor value by default)
    color : optional
        Variable to split & colour dots by (cell group by default)
    groups : optional
        Subset of groups to consider
    group_label : optional
        Sample (cell) metadata column to be used as group assignment ('group' by default)
    violins : optional
        Boolean value if to add violin plots
    dots : optional
        Boolean value if to add dots to the plots
    zero_line : optional
        Boolean values if to add Z=0 line
    linewidth : optional
        Linewidth argument for dots (default is 0)
    zero_linewidth : optional
        Linewidth argument for the zero line (default is 1)
    size : optional
        Size argument for dots (ms for plot, s for jointplot and scatterplot; default is 5)
    legend : optional bool
        If to show the legend (e.g. colours matching groups)
    legend_prop : optional
        The font properties of the legend
    palette : optional
        cmap describing colours, default is None (cubehelix)
        Example palette: seaborn.cubehelix_palette(8, start=.5, rot=-.75. as_cmap=True)
    alpha : optional
        Dots opacity
    violins_alpha : optional
        Violins opacity
    ncols : optional
        Number of columns if multiple colours are defined (4 by default)
    sharex: optional
        Common X axis across plots on the grid
    sharey: optional
        Common Y axis across plots on the grid
    """

    # Process input arguments
    if group_label == "group" and color is None:
        color = "group"
    color_vars = maybe_factor_indices_to_factors(_make_iterable(color))

    assert violins or dots, "Either violins=True or dots=True"

    # Get factors
    z = model.get_factors(factors=factors, df=True)
    z = z.rename_axis("sample").reset_index()
    # Make the table long for plotting
    z = z.melt(id_vars="sample", var_name="factor", value_name="value")

    # Add group and colour information
    vars = [group_label, *color_vars]
    if any([not (not (i)) for i in vars]):
        meta = model.fetch_values(variables=vars)
        z = z.set_index("sample").join(meta).reset_index()

    # Subset groups (incl. custom groups of samples)
    if group_label and groups is not None:
        z = z[z[group_label].isin(groups)]

    z["factor_idx"] = z.factor.str.lstrip("Factor").astype(int)
    z = z.sort_values(by="factor_idx")

    modifier = None
    if dots:

        def modifier(
            data,
            ax,
            split_var,
            color_var,
            alpha=alpha,
        ):
            # Add dots
            sns.stripplot(
                data=z,
                dodge=True,
                size=size,
                x="factor",
                y="value",
                hue=color_var,
                alpha=alpha,
                linewidth=linewidth,
                ax=ax,
            )

        modifier = partial(modifier, data=z)

    plot = partial(
        sns.violinplot,
        inner=None,
        s=size,
    )

    g = _plot_grid(
        plot,
        z,
        x="factor",
        y="value",
        color=color,
        zero_line_x=False,
        zero_line_y=zero_line,
        linewidth=linewidth,
        zero_linewidth=zero_linewidth,
        size=size,
        legend=legend,
        legend_prop=legend_prop,
        palette=palette,
        ncols=ncols,
        sharex=sharex,
        sharey=sharey,
        modifier=modifier,
        **kwargs,
    )

    # Adjust violins alpha
    if violins:
        if violins_alpha:
            for path in g.collections:
                if path.__class__.__name__ == "PolyCollection":
                    path.set_alpha(violins_alpha)
    if violins is None or not violins or violins_alpha == 0:
        n_plots = len(g.collections)
        i = 0
        while i < len(g.collections):
            path = g.collections[i]
            if path.__class__.__name__ == "PolyCollection":
                path.remove()
            else:
                i += 1

    return g


def plot_factors_umap(
    model: mofa_model,
    factors: Optional[Union[int, List[int]]] = None,
    groups=None,
    group_label: Optional[str] = None,
    color=None,
    linewidth=0,
    size=20,
    n_neighbors=10,
    spread=1,
    min_dist=0.5,
    random_state=None,
    umap_kwargs={},
    legend=True,
    legend_prop=None,
    palette=None,
    ncols=4,
    sharex=False,
    sharey=False,
    **kwargs,
):
    """
    Plot UMAP on factor values.

    Parameters
    ----------
    model : mofa_model
        Factor model
    factors : optional
        Index of a factor (or indices of factors) to use (all factors by default)
    groups : optional
        Subset of groups to consider
    group_label : optional
        Sample (cell) metadata column to be used as group assignment
    color : optional
        Grouping variable by default, alternatively a feature name can be provided
    linewidth : optional
        Linewidth argument for dots (default is 0)
    size : optional
        Size argument for dots (ms for plot, s for jointplot and scatterplot; default is 5)
    n_neighbors : optional
        n_neighbors parameter for UMAP
    spread : optional
        spread parameter for UMAP
    random_state : optional
        random_state parameter for UMAP
    umap_kwargs : optional
        Additional arguments to umap.UMAP()
    legend : optional bool
        If to show the legend (e.g. colours matching groups)
    legend_prop : optional
        The font properties of the legend
    palette : optional
        cmap describing colours, default is None (cubehelix)
        Example palette: seaborn.cubehelix_palette(8, start=.5, rot=-.75. as_cmap=True)
    ncols : optional
        Number of columns if multiple colours are defined (4 by default)
    """

    # Process input arguments
    if group_label == "group" and not color:
        color = "group"
    color_vars = maybe_factor_indices_to_factors(_make_iterable(color))

    # Check if UMAP has be pre-computed
    get_umap_cols = lambda m: np.where(
        [s.startswith("UMAP") for s in m.samples_metadata.columns.values]
    )[0]
    umap_cols = get_umap_cols(model)
    if len(umap_cols) == 0:
        print("Computing UMAP coordinates...")
        model.run_umap(
            factors=factors,
            n_neighbors=n_neighbors,
            spread=spread,
            min_dist=min_dist,
            random_state=random_state,
        )
        umap_cols = get_umap_cols(model)

    embedding = model.samples_metadata.iloc[:, umap_cols]

    x, y, *_ = embedding.columns

    # Add group and colour information
    vars = [group_label, *color_vars]
    vars = [v for v in vars if v not in embedding.columns.values]
    if any([not (not (i)) for i in vars]):
        meta = model.fetch_values(variables=vars)
        embedding = embedding.rename_axis("sample").reset_index()
        embedding = embedding.set_index("sample").join(meta).reset_index()

    # Subset groups (incl. custom groups of samples)
    if group_label and groups is not None:
        embedding = embedding[embedding[group_label].isin(groups)]

    plot = partial(
        sns.scatterplot,
        s=size,
    )

    g = _plot_grid(
        plot,
        embedding,
        x=x,
        y=y,
        color=color,
        zero_line_x=False,
        zero_line_y=False,
        linewidth=linewidth,
        size=size,
        legend=legend,
        legend_prop=legend_prop,
        palette=palette,
        ncols=ncols,
        sharex=sharex,
        sharey=sharey,
        **kwargs,
    )

    return g


# The following functions do not use the grid interface


def plot_factors_matrix(
    model: mofa_model,
    factors: Optional[Union[int, List[int], str, List[str]]] = None,
    group_label: Optional[str] = None,
    groups: Optional[Union[int, List[int], str, List[str]]] = None,
    agg="mean",
    cmap="viridis",
    vmin=None,
    vmax=None,
    **kwargs,
):
    """
    Average factor value per group and plot it as a heatmap

    Parameters
    ----------
    model : mofa_model
        Factor model
    factors : optional
        Factor idices or names (all factors by default)
    group_label : optional
        Sample (cell) metadata column to be used as group assignment
    groups : optional
        Group indices or names (all groups by default)
    avg : optional
        Aggregation function to average factor values per group (mean by default)
    cmap : optional
        Heatmap cmap argument ("viridis" by default)
    """

    z = model.get_factors(factors=factors, groups=groups, df=True)
    z = z.rename_axis("sample").reset_index()
    # Make the table long for plotting
    z = z.melt(id_vars="sample", var_name="factor", value_name="value")

    # Assign a group to every sample (cell) if it is provided
    if group_label is None:
        group_label = "group"
    groups_df = model.samples_metadata.loc[:, [group_label]]

    groups_df.rename(columns={groups_df.columns[0]: "group"}, inplace=True)

    # Add group information for samples (cells)
    z = z.set_index("sample").join(groups_df).reset_index()

    z = (
        z.groupby(["group", "factor"])
        .agg({"value": agg})
        .reset_index()
        .pivot(index="factor", columns="group", values="value")
    )

    z.index = z.index.astype("category")
    z.index = z.index.reorder_categories(
        sorted(z.index.categories, key=lambda x: int(x.split("Factor")[1]))
    )
    z = z.sort_values("factor", ascending=False)

    ax = sns.heatmap(z, cmap=cmap, vmin=vmin, vmax=vmax, **kwargs)
    ax.set(ylabel="Factor", xlabel="Group")
    ax.set_yticklabels(ax.yaxis.get_ticklabels(), rotation=0)

    return ax


def plot_factors_dotplot(
    model: mofa_model,
    factors: Optional[Union[int, List[int], str, List[str]]] = None,
    group_label: Optional[str] = None,
    groups: Optional[Union[int, List[int], str, List[str]]] = None,
    palette=None,
    vmin=None,
    vmax=None,
    size=100,
    xticklabels_size=8,
    yticklabels_size=8,
    **kwargs,
):
    """
    Average factor value per group and plot it as a heatmap.
    Colour indicates mean factor value
    and size relates to its variance.

    Parameters
    ----------
    model : mofa_model
        Factor model
    factors : optional
        Factor idices or names (all factors by default)
    group_label : optional
        Sample (cell) metadata column to be used as group assignment
    groups : optional
        Group indices or names (all groups by default)
    palette : optional
        Colour palette (blue-white-red by default)
    vmin : optional
        Min color value
    vmax : optional
        Max colour value
    size : optional
        Dot plot size scaling coefficient
    xticklabels_size : optional
        Font size for features labels (default is 10)
    yticklabels_size : optional
        Font size for factors labels (default is None)
    """

    if palette is None:
        palette = sns.diverging_palette(240, 10, sep=3, center="light", as_cmap=True)

    z = model.get_factors(factors=factors, groups=groups, df=True)
    z = z.rename_axis("sample").reset_index()
    # Make the table long for plotting
    z = z.melt(id_vars="sample", var_name="factor", value_name="value")

    # Assign a group to every sample (cell) if it is provided
    if group_label is None:
        group_label = "group"
    groups_df = model.samples_metadata.loc[:, [group_label]]

    groups_df.rename(columns={groups_df.columns[0]: "group"}, inplace=True)

    # Add group information for samples (cells)
    z = z.set_index("sample").join(groups_df).reset_index()

    z = z.groupby(["group", "factor"]).agg({"value": ["mean", "var"]}).reset_index()
    z.columns = ["group", "factor", "value_mean", "value_var"]
    z["abs_mean"] = np.abs(z.value_mean)

    z.factor = z.factor.astype("category")
    z.factor = z.factor.cat.reorder_categories(
        sorted(z.factor.cat.categories, key=lambda x: int(x.split("Factor")[1]))
    )
    z = z.sort_values("factor")

    # Fix group order
    groups = (
        z.pivot(columns="group", index="factor", values="abs_mean")
        .idxmax(axis=0)
        .sort_values(key=lambda x: x.str.split("Factor").str[1].astype(int))
        .index.values
    )
    z.group = z.group.astype("category")
    z.group = z.group.cat.reorder_categories(np.flip(groups))

    z = z.sort_values(["factor", "group"])

    # Normalise colour map
    abs_mean_max = max(abs(z.value_mean.min()), z.value_mean.max())
    norm = plt.Normalize(-abs_mean_max, abs_mean_max)
    # norm = plt.Normalize(z.value_mean.min(), z.value_mean.max())
    sm = plt.cm.ScalarMappable(cmap=palette, norm=norm)
    # from matplotlib.colors import DivergingNorm
    # divnorm = DivergingNorm(vmin=-abs_mean, vcenter=0, vmax=abs_mean)
    sm.set_array([])

    # Scale dot size
    z.value_var = z.value_var / z.value_var.max() * size

    # Create the plot
    fig, ax = plt.subplots()
    g = plt.scatter(
        x=z.factor,
        y=z.group,
        c=z.value_mean,
        s=z.value_var,
        cmap=palette,
        norm=norm,
        vmin=vmin,
        vmax=vmax,
        **kwargs,
    )

    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)

    try:
        g.figure.colorbar(sm)
        g.get_legend().remove()
    except Exception:
        warn("Cannot make a proper colorbar")

    plt.xticks(rotation=90, size=xticklabels_size)
    plt.yticks(size=yticklabels_size)

    plt.ylabel(group_label)

    return g


def plot_factors_correlation(
    model: mofa_model,
    factors: Optional[Union[int, List[int]]] = None,
    groups: Optional[Union[int, List[int], str, List[str]]] = None,
    covariates=None,
    pvalues=False,
    linewidths=0,
    diag=False,
    cmap=None,
    square=True,
    **kwargs,
):
    """
    Plot correlation of factors and, if provided, covariates

    Parameters
    ----------
    model : mofa_model
        Factor model
    factors : optional
        Index of a factor (or indices of factors) to use (all factors by default)
    groups : optional
        Subset of groups to consider
    covarites : optional
        A vector, a matrix, or a data frame with covariates (one per column)
    pvalues
        Plot BH-adjusted p-values instead of correlation coefficient
    linewidths : optional
        Heatmap linewidths argument (default is 0)
    diag : optional
        If to only plot lower triangle of the correlation matrix (False by default)
    cmap : optional
        Heatmap cmap argument
    square : optional
        Heatmap square argument (True by default)
    """

    z = model.get_factors(factors=factors, groups=groups)

    # pearsonr returns (r, pvalue)
    value_index = 1 if pvalues else 0

    if covariates is not None:
        n_cov = covariates.shape[1]
        # Transform a vector to a data frane
        # Also ransform matrices and ndarrays to a data frame
        if len(covariates.shape) == 1 or not isinstance(covariates, pd.DataFrame):
            covariates = pd.DataFrame(covariates)
        corr = np.ndarray(shape=(z.shape[1], n_cov))
        for i in range(corr.shape[0]):
            for j in range(corr.shape[1]):
                corr[i, j] = pearsonr(z[:, i], covariates.iloc[:, j])[value_index]
    else:
        # Inter-factor correlations
        corr = np.ndarray(shape=(z.shape[1], z.shape[1]))
        for i in range(corr.shape[0]):
            for j in range(corr.shape[1]):
                corr[i, j] = pearsonr(z[:, i], z[:, j])[value_index]

    if pvalues:
        corr = padjust_fdr_2d(corr)

        corr = -np.log10(corr)

        if np.sum(np.isinf(corr)) > 0:
            warn("Some p-values are 0, these values will be capped.")
            corr[np.isinf(corr)] = np.ceil(corr[~np.isinf(corr)].max() * 10)

    mask = None
    if diag:
        # Generate a mask for the upper triangle
        mask = np.triu(np.ones_like(corr, dtype=np.bool))

    # Set up the matplotlib figure
    f, ax = plt.subplots()

    if cmap is None:
        # Generate a custom diverging colormap
        cmap = "Reds" if pvalues else sns.diverging_palette(220, 10, as_cmap=True)

    # Generate labels for the heatmap
    if factors is None:
        factors = range(z.shape[1])
    fnames = [f"Factor{fi+1}" if isinstance(fi, int) else fi for fi in factors]
    if covariates is not None:
        if isinstance(covariates, pd.DataFrame):
            cnames = covariates.columns.values
        else:
            cnames = [f"Covar{ci+1}" for ci in covariates.shape[1]]
        xticklabels = cnames
        yticklabels = fnames
    else:
        xticklabels = fnames
        yticklabels = fnames

    center = 0 if not pvalues else None
    cbar_kws = {"shrink": 0.5}
    cbar_kws["label"] = (
        "Correlation coefficient" if not pvalues else "-log10(adjusted p-value)"
    )

    # Draw the heatmap with the mask and correct aspect ratio
    g = sns.heatmap(
        corr,
        cmap=cmap,
        mask=mask,
        center=center,
        square=True,
        linewidths=0.5,
        xticklabels=xticklabels,
        yticklabels=yticklabels,
        cbar_kws=cbar_kws,
        **kwargs,
    )

    g.set_yticklabels(g.yaxis.get_ticklabels(), rotation=0)

    return g


def plot_factors_covariates_correlation(
    model: mofa_model,
    covariates: Union[np.ndarray, np.matrix, pd.DataFrame],
    pvalues: bool = False,
    **kwargs,
):
    """
    Plot correlation of factors and covariates

    Parameters
    ----------
    model : mofa_model
        Factor model
    covarites
        A vector, a matrix, or a data frame with covariates (one per column)
    pvalues
        Plot BH-adjusted p-values instead of correlation coefficient
    **kwargs
        Other arguments to plot_factors_correlation
    """
    return plot_factors_correlation(
        model, covariates=covariates, pvalues=pvalues, square=False, **kwargs
    )


### Projection ###


def plot_projection(
    model: mofa_model,
    data,
    data_name: str = "projected_data",
    view: Union[str, int] = None,
    with_orig: bool = False,
    x="Factor1",
    y="Factor2",
    groups=None,
    groups_df=None,
    color=None,
    linewidth=0,
    size=20,
    legend=False,
    legend_loc="best",
    legend_prop=None,
    feature_intersection=False,
    **kwargs,
):
    """
    Project new data onto the factor space of the model.

    For the projection, a pseudo-inverse of the weights matrix is calculated
    and its product with the provided data matrix is calculated.

    Parameters
    ----------
    model : mofa_model
        Factor model
    data
        Numpy array or Pandas DataFrame with the data matching the number of features
    data_name : optional
        A name for the projected dataset ("projected_data" by default)
    view : optional
        A view of the model to consider (first view by default)
    with_orig : optional
        Boolean value if to plot data from the model (False by default)
    x : optional
        Factor to plot along X axis (Factor1 by default)
    y : optional
        Factor to plot along Y axis (Factor2 by default)
    groups : optional
        Subset of groups to consider
    groups_df : optional pd.DataFrame
        Data frame with samples (cells) as index and first column as group assignment
    color : optional
        A feature name to colour the dots by its expression
    linewidth : optional
        Linewidth argument for dots (default is 0)
    size : optional
        Size argument for dots (ms for plot, s for jointplot and scatterplot; default is 5)
    legend : optional bool
        If to show the legend (e.g. colours matching groups)
    legend_loc : optional
        Legend location (e.g. 'upper left', 'center', or 'best')
    legend_prop : optional
        The font properties of the legend
    feature_intersection : optional
        Feature intersection flag for project_data
    """
    zpred = model.project_data(
        data=data,
        view=view,
        factors=[x, y],
        df=True,
        feature_intersection=feature_intersection,
    )
    zpred.columns = ["x", "y"]

    # Get and prepare Z matrix from the model if required
    if with_orig:
        z = model.get_factors(factors=[x, y], groups=groups, df=True)
        z.columns = ["x", "y"]

        # Assign a group to every sample (cell) if it is provided
        if groups_df is None:
            groups_df = model.get_samples().set_index("sample")

        z = z.rename_axis("sample").reset_index()
        z = z.set_index("sample").join(groups_df).reset_index()
        grouping_var = groups_df.columns[0]

        # Assign colour to every sample (cell) if colouring by feature expression
        if color is None:
            color_var = grouping_var
        else:
            color_var = color
            color_df = model.get_data(features=color, df=True)
            z = z.set_index("sample").join(color_df).reset_index()
            z = z.sort_values(color_var)
    else:
        grouping_var = "group"
    zpred[grouping_var] = data_name

    # Assign colour to every sample (cell) in the new data if colouring by feature expression
    if color is None:
        color_var = grouping_var
    else:
        if isinstance(data, pd.DataFrame):
            color_var = color
            color_df = data.loc[:, [color]]
            zpred = zpred.join(color_df).reset_index()
            zpred = zpred.sort_values(color_var)

    if with_orig:
        z = z.append(zpred)
    else:
        z = zpred

    # Define plot axes labels
    x_factor_label = f"Factor{x+1}" if isinstance(x, int) else x
    y_factor_label = f"Factor{y+1}" if isinstance(y, int) else y

    # Set default colour to black if none set
    if "c" not in kwargs and "color" not in kwargs:
        kwargs["color"] = "black"

    g = sns.scatterplot(
        x="x",
        y="y",
        data=z,
        linewidth=linewidth,
        s=size,
        hue=color_var,
        legend=legend,
        **kwargs,
    )
    sns.despine(offset=10, trim=True, ax=g)
    g.set(xlabel=f"{x_factor_label} value", ylabel=f"{y_factor_label} value")

    return g



import pandas as pd
import numpy as np
from typing import Union, List, Optional, Iterable
import h5py

#######################
## Loading metadata  ##
#######################


def _load_samples_metadata(model):
    samples_metadata = pd.DataFrame(
        [
            [cell, group]
            for group, cell_list in model.samples.items()
            for cell in cell_list
        ],
        columns=["sample", "group"],
    )
    if "samples_metadata" in model.model:
        if len(list(model.model["samples_metadata"][model.groups[0]].keys())) > 0:
            _samples_metadata = pd.concat(
                [
                    pd.concat(
                        [
                            pd.Series(model.model["samples_metadata"][g][k])
                            for k in model.model["samples_metadata"][g].keys()
                        ],
                        axis=1,
                    )
                    for g in model.groups
                ],
                axis=0,
            )
            _samples_metadata.columns = list(
                model.model["samples_metadata"][model.groups[0]].keys()
            )

            if "group" in _samples_metadata.columns:
                del _samples_metadata["group"]
            if "sample" in _samples_metadata.columns:
                del _samples_metadata["sample"]

            samples_metadata = pd.concat(
                [
                    samples_metadata.reset_index(drop=True),
                    _samples_metadata.reset_index(drop=True),
                ],
                axis=1,
            )

            # Decode objects as UTF-8 strings
            for column in samples_metadata.columns:
                if samples_metadata[column].dtype == "object":
                    try:
                        samples_metadata[column] = [
                            i.decode() for i in samples_metadata[column].values
                        ]
                    except (UnicodeDecodeError, AttributeError):
                        pass

    samples_metadata = samples_metadata.set_index("sample")
    return samples_metadata


def _load_features_metadata(model):
    features_metadata = pd.DataFrame(
        [
            [feature, view]
            for view, feature_list in model.features.items()
            for feature in feature_list
        ],
        columns=["feature", "view"],
    )
    if "features_metadata" in model.model:
        if len(list(model.model["features_metadata"][model.views[0]].keys())) > 0:
            features_metadata_dict = {
                m: pd.concat(
                    [
                        pd.Series(model.model["features_metadata"][m][k])
                        for k in model.model["features_metadata"][m].keys()
                    ],
                    axis=1,
                )
                for m in model.views
            }

            for m in features_metadata_dict.keys():
                features_metadata_dict[m].columns = list(
                    model.model["features_metadata"][m].keys()
                )

            _features_metadata = pd.concat(features_metadata_dict, axis=0)

            if "view" in _features_metadata.columns:
                del _features_metadata["view"]
            if "feature" in _features_metadata.columns:
                del _features_metadata["feature"]

            features_metadata = pd.concat(
                [
                    features_metadata.reset_index(drop=True),
                    _features_metadata.reset_index(drop=True),
                ],
                axis=1,
            )

            # Decode objects as UTF-8 strings
            for column in features_metadata.columns:
                if features_metadata[column].dtype == "object":
                    try:
                        features_metadata[column] = [
                            i.decode() for i in features_metadata[column].values
                        ]
                    except (UnicodeDecodeError, AttributeError):
                        pass

    features_metadata = features_metadata.set_index("feature")
    return features_metadata


def _load_covariates(model):
    # Covariates names
    cov_names = None
    if "covariates" in model.model:
        cov_root = model.model["covariates"]
        if (type(cov_root) == h5py.Group) and ("covariates" in cov_root):
            cov_root = cov_root["covariates"]
        cov_names = np.array(cov_root)
        try:
            cov_names = [i.decode() for i in cov_names]
        except (UnicodeDecodeError, AttributeError):
            pass
    else:
        if "cov_samples" in model.model:
            cov_len = model.model["cov_samples"][model.groups[0]].shape[-1]
            cov_names = [f"Covariates{i+1}" for i in range(cov_len)]
        else:
            return None, None

    # Covariates values
    samples_covariates = pd.DataFrame(
        [
            [cell, group]
            for group, cell_list in model.samples.items()
            for cell in cell_list
        ],
        columns=["sample", "group"],
    )

    for attr in ["cov_samples", "cov_samples_transformed"]:
        if attr in model.model:
            if len(model.model[attr].keys()) > 0:  # groups are not empty
                attr_covariates = np.concatenate(
                    [np.array(model.model[attr][g]) for g in model.groups]
                )
                attr_covariates = pd.DataFrame(attr_covariates)
                if attr == "cov_samples":
                    attr_covariates.columns = cov_names
                elif attr == "cov_samples_transformed":
                    attr_covariates.columns = [f"{n}_transformed" for n in cov_names]
                else:
                    attr_covariates.columns = [
                        f"{attr}{i+1}" for i in range(attr_covariates.shape[-1])
                    ]
                samples_covariates = pd.concat(
                    [samples_covariates, attr_covariates], axis=1
                )

    # Covariates are None when there is only sample name and group name
    if len(samples_covariates.columns) == 2:
        return None, None

    samples_covariates = samples_covariates.set_index("sample")

    return cov_names, samples_covariates


#######################
#### Reading HDF5 #####
#######################


def _read_simple(data):
    if isinstance(data, h5py.Group):
        return {k: np.array(v) for k, v in data.items()}
    elif isinstance(data, h5py.Dataset):
        return np.array(data)
    else:
        raise ValueError("Attempting to read neither Group nor Dataset from HDF5 file")


#######################
## Utility functions ##
#######################

### Factors


def factor_indices_to_factors(x):
    if x is None:
        return None
    elif isinstance(x, int):
        x = [x]
    return [f"Factor{i+1}" for i in x]


def maybe_factor_indices_to_factors(x):
    if x is None:
        return None
    elif isinstance(x, int):
        return f"Factor{x+1}"
    elif isinstance(x, str):
        return x
    return [f"Factor{i+1}" if isinstance(i, int) else i for i in x]


def _is_iter(x):
    return isinstance(x, Iterable) and not isinstance(x, str)


def _make_iterable(x):
    if not _is_iter(x):  # e.g. int
        x = [x]
    return x


### R2 statistics


def calculate_r2(Z, W, Y):
    a = np.nansum((Y - Z.T.dot(W)) ** 2.0)
    b = np.nansum(Y ** 2)
    r2 = (1.0 - a / b) * 100
    return r2


def padjust_fdr(xs):
    """
    Adjust p-values using the BH procedure
    """
    from scipy.stats import rankdata

    ranked_p_values = rankdata(xs)
    fdr = xs * len(xs) / ranked_p_values
    fdr[fdr > 1] = 1
    return fdr


def padjust_fdr_2d(mx):
    """
    Adjust p-values in a matrix using the BH procedure
    """
    from scipy.stats import rankdata

    ranked_p_values = rankdata(mx).reshape((-1, mx.shape[1]))
    fdr = mx * mx.shape[0] * mx.shape[1] / ranked_p_values
    fdr[fdr > 1] = 1
    return fdr



from .core import mofa_model
from .utils import *

import sys
from warnings import warn
from typing import Union, Optional, List, Iterable, Sequence
from functools import partial

import numpy as np
from scipy.stats import pearsonr
import pandas as pd
from pandas.api.types import is_numeric_dtype
import matplotlib.pyplot as plt
from matplotlib import rcParams
import seaborn as sns

from .utils import maybe_factor_indices_to_factors, _make_iterable, _is_iter
from .plot_utils import _plot_grid

from .plot_factors import *
from .plot_weights import *
from .plot_variance import *
from .plot_mefisto import *



from .core import mofa_model
from .utils import *

from typing import Union, Optional, List, Iterable, Sequence
from functools import partial

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib import rcParams
import seaborn as sns

from .utils import _make_iterable
from .plot_utils import _plot_grid


def plot_interpolated_factors(
    model: mofa_model,
    factors: Union[int, List[int]] = None,
    groups=None,
    only_mean: bool = False,
    show_observed: bool = True,
    color=None,
    alpha=0.3,
    zero_line=False,
    linewidth=2,
    dot_linewidth=0,
    zero_linewidth=1,
    size=20,
    legend=True,
    legend_prop=None,
    palette=None,
    ncols=4,
    sharex=False,
    sharey=False,
    **kwargs,
):
    """
    Plot samples features such as factor values,
    samples metadata or covariates

    Parameters
    ----------
    model : mofa_model
        Factor model
    factors : optional
        Index of a factor (or indices of factors) to use (all factors by default)
    groups : optional
        Subset of groups to consider
    only_mean : optional
        If not to plot confidence for the interpolated values (False by default)
    show_observed : optional
        If not to plot oberved factor values along the transformed covariate (True by default)
    group_label : optional
        Sample (cell) metadata column to be used as group assignment ('group' by default)
    color : optional
        Grouping variable by default, alternatively a feature name can be provided (when no kde).
        If a list of features is provided, they will be plot on one figure.
        Use palette argument to provide a colour map.
    alpha : optional
        Opacity when plotting confidence for the interpolated values (when only_mean=False)
    zero_line : optional
        Boolean values if to add Z=0 line
    linewidth : optional
        Linewidth argument for lines (default is 2)
    dot_linewidth : optional
        Linewidth argument for dots (default is 0)
    zero_linewidth : optional
        Linewidth argument for the zero line (default is 1)
    size : optional
        Size argument for dots (ms for plot, s for jointplot and scatterplot; default is 5)
    legend : optional bool
        If to show the legend (e.g. colours matching groups)
    legend_prop : optional
        The font properties of the legend
    palette : optional
        cmap describing colours, default is None (cubehelix)
        Example palette: seaborn.cubehelix_palette(8, start=.5, rot=-.75. as_cmap=True)
    ncols : optional
        Number of columns if multiple colours are defined (4 by default)
    sharex: optional
        Common X axis across plots on the grid
    sharey: optional
        Common Y axis across plots on the grid
    """

    # Process input arguments
    if color is None:
        color = "group"
    if color != "group":
        raise ValueError(
            "Only colouring by group is supported when plotting interpolated factors"
        )

    factors, factor_indices = model._check_factors(factors)

    # Get factors
    zi = model.get_interpolated_factors(
        factors=factors, df_long=True
    )  # this includes the group

    new_values_dim = model.interpolated_factors["new_values"].shape[1]
    if new_values_dim == 1 and len(model.covariates_names) == 1:
        new_value = f"{model.covariates_names[0]}_transformed"
    else:
        # If the new values are multi-dimensional, address them by index
        # since multi-dimensional plots are not supported
        new_value = "new_value"

    # Subset groups
    if groups is not None:
        zi = zi[zi["group"].isin(_make_iterable(groups))]

    zi_mean = (
        zi.pivot(
            index=["new_sample", new_value, "group"], columns="factor", values="mean"
        )
        .reset_index()
        .rename_axis(None, axis=1)
        .sort_values([new_value, "group"])
    )
    zi_var = (
        zi.pivot(
            index=["new_sample", new_value, "group"],
            columns="factor",
            values="variance",
        )
        .reset_index()
        .rename_axis(None, axis=1)
        .sort_values([new_value, "group"])
    )

    z_observed = None
    if show_observed:
        covs = [f"{v}_transformed" for v in model.covariates_names]
        if len(covs) > 1:
            raise NotImplemented(
                "Only data with a single covariate is currently supported"
            )
        z_observed = model.fetch_values([*factors, covs[0], "group"]).sort_values(
            ["group"]
        )
        # Subset groups
        if groups is not None:
            z_observed = z_observed[z_observed["group"].isin(_make_iterable(groups))]

    plot = partial(
        sns.lineplot,
    )

    modifier = None
    if (not only_mean) or show_observed:

        def modifier(
            data,
            ax,
            split_var,
            color_var,
            alpha=alpha,
            show_observed=show_observed,
            only_mean=only_mean,
        ):
            m, v = data["mean"], data["var"]

            # Add confidence intervals
            if not only_mean:
                get_conf = lambda f, mean, var: f(mean, 1.96 * np.sqrt(var))
                for group in m[color_var].unique():
                    m_g = m[m[color_var] == group]
                    v_g = v[v[color_var] == group]
                    g_mean, g_var = m_g[split_var].values, v_g[split_var].values
                    ax.fill_between(
                        m_g[new_value],
                        get_conf(np.subtract, g_mean, g_var),
                        get_conf(np.add, g_mean, g_var),
                        alpha=alpha,
                    )

            # Add dots
            if show_observed:
                sns.scatterplot(
                    data=data["observed"],
                    x=new_value,
                    y=split_var,
                    hue=color_var,
                    linewidth=dot_linewidth,
                    ax=ax,
                )

        modifier = partial(
            modifier, data={"mean": zi_mean, "var": zi_var, "observed": z_observed}
        )

    g = _plot_grid(
        plot,
        zi_mean,
        x=new_value,
        y=factors,
        color=color,
        zero_line_x=False,
        zero_line_y=zero_line,
        linewidth=linewidth,
        zero_linewidth=zero_linewidth,
        legend=legend,
        legend_prop=legend_prop,
        palette=palette,
        ncols=ncols,
        sharex=sharex,
        sharey=sharey,
        modifier=modifier,
        **kwargs,
    )

    return g


### MEFISTO ###


def plot_group_kernel(
    model, groups=None, factors=None, palette=None, vmin=-1, vmax=1, ncols=4, **kwargs
):

    z = model.get_factors(factors=factors, groups=groups)
    factor_indices, factors = model._check_factors(factors, unique=True)

    groups = model._check_groups(groups)
    all_groups = np.array(model.groups)
    group_indices = [np.where(all_groups == gr)[0][0] for gr in groups]

    # Get group kernels
    Kgs = model.get_group_kernel()[factor_indices, :, :][:, group_indices, :][
        :, :, group_indices
    ]

    df_list = [
        pd.DataFrame(Kgs[i], index=groups, columns=groups) for i in range(Kgs.shape[0])
    ]

    if palette is None:
        palette = "RdBu_r"

    # Only use clustermap when there's one factors to plot
    # since it creates its own figure
    if len(factors) == 1:
        g = sns.clustermap(df_list[0], cmap=palette, vmin=vmin, vmax=vmax, **kwargs)
        g.ax_heatmap.set(ylabel="Group", xlabel="Group", title=factors[0])
        g.ax_heatmap.set_yticklabels(g.ax_heatmap.yaxis.get_ticklabels(), rotation=0)

        sns.despine(offset=10, trim=True, ax=g.ax_heatmap)

    else:
        # Figure out rows & columns for the grid with plots
        ncols = min(ncols, len(factors))
        nrows = int(np.ceil(len(factors) / ncols))
        fig, axes = plt.subplots(
            nrows,
            ncols,
            figsize=(
                ncols * rcParams["figure.figsize"][0],
                nrows * rcParams["figure.figsize"][1],
            ),
        )

        if ncols == 1:
            axes = np.array(axes).reshape(-1, 1)
        if nrows == 1:
            axes = np.array(axes).reshape(1, -1)

        for i, factor in enumerate(factors):

            ri = i // ncols
            ci = i % ncols

            g = sns.heatmap(
                df_list[i],
                square=True,
                cmap=palette,
                vmin=vmin,
                vmax=vmax,
                ax=axes[ri, ci],
                **kwargs,
            )
            g.set(ylabel="Group", xlabel="Group", title=factor)
            g.set_yticklabels(g.yaxis.get_ticklabels(), rotation=0)

            sns.despine(offset=10, trim=True, ax=g)

        # Remove unused axes
        for i in range(len(factors), ncols * nrows):
            ri = i // ncols
            ci = i % ncols
            fig.delaxes(axes[ri, ci])

    plt.tight_layout()

    return g


def plot_sharedness(
    model, groups=None, factors=None, color="#B8CF87", return_data=False, **kwargs
):
    GROUPS_MSG = "Multiple groups are required to determine sharedness"
    assert model.ngroups > 1, GROUPS_MSG
    if groups is not None:
        assert not isinstance(groups, str) and len(groups) > 1, GROUPS_MSG

    z = model.get_factors(factors=factors, groups=groups)
    factor_indices, factors = model._check_factors(factors, unique=True)

    groups = model._check_groups(groups)
    all_groups = np.array(model.groups)
    group_indices = [np.where(all_groups == gr)[0][0] for gr in groups]

    # Get group kernels
    Kgs = model.get_group_kernel()[factor_indices, :, :][:, group_indices, :][
        :, :, group_indices
    ]

    # Calculate distance
    gr = np.array(
        [
            np.abs(Kgs[i, :, :])[np.tril(Kgs[i, :, :], -1).astype(bool)].mean()
            for i in range(Kgs.shape[0])
        ]
    )

    df = pd.DataFrame({"factor": factors, "shared": gr, "non_shared": 1 - gr})

    if return_data:
        return df

    sns.barplot(data=df, y="factor", x="shared", color="lightgrey")
    g = sns.barplot(data=df, color=color, y="factor", x="shared", **kwargs)

    g.set(xlabel="Sharedness", ylabel="Factor")

    sns.despine(offset=10, trim=True, ax=g)

    plt.tight_layout()

    return g


def plot_smoothness(model, factors=None, color="#5F9EA0", return_data=False, **kwargs):
    z = model.get_factors(factors=factors)
    factor_indices, factors = model._check_factors(factors, unique=True)

    # Get scales
    scales = np.array(model.model["training_stats"]["scales"])

    df = pd.DataFrame({"factor": factors, "smooth": scales, "non_smooth": 1 - scales})

    if return_data:
        return df

    sns.barplot(data=df, y="factor", x="smooth", color="lightgrey")
    g = sns.barplot(data=df, color=color, y="factor", x="smooth", **kwargs)

    g.set(xlabel="Smoothness", ylabel="Factor")

    sns.despine(offset=10, trim=True, ax=g)

    plt.tight_layout()

    return g



import h5py
import numpy as np
import pandas as pd

import sys
from os import path
from typing import Union, List, Optional
from collections.abc import Iterable
import warnings

from .utils import (
    _load_samples_metadata,
    _load_features_metadata,
    _load_covariates,
    _read_simple,
)
from .utils import *


class mofa_model:
    """Class around HDF5-based model on disk.

    This class is a thin wrapper for the HDF5 file where the trained MOFA+ model is stored.
    It also provides utility functions to get factors, weights, features, and samples (cells) info
    in the form of Pandas dataframes, and data as a NumPy array.
    """

    def __init__(self, filepath, mode="r"):
        self.filepath = filepath
        self.filename = path.basename(filepath)
        self.model = h5py.File(filepath, mode)

        # Define samples
        self.samples = {
            g: np.array(self.model["samples"][g]).astype("str")
            for g in self.model["samples"]
        }

        # Define features
        self.features = {
            m: np.array(self.model["features"][m]).astype("str")
            for m in self.model["features"]
        }

        # Define groups
        self.groups = (
            list(np.array(self.model["groups"]["groups"]).astype(str))
            if "groups" in self.model
            else list(self.model["samples"].keys())
        )

        # Define views
        self.views = (
            list(np.array(self.model["views"]["views"]).astype(str))
            if "views" in self.model
            else list(self.model["features"].keys())
        )

        # Define data and epxectations
        self.data = self.model["data"] if "data" in self.model else None
        self.expectations = self.model["expectations"]
        self.factors = self.model["expectations"]["Z"]
        self.weights = self.model["expectations"]["W"]

        # Define dimensionalities
        if self.data is not None:
            self.shape = (
                sum(self.data[self.views[0]][group].shape[0] for group in self.groups),
                sum(self.data[view][self.groups[0]].shape[1] for view in self.views),
            )
        else:
            self.shape = (
                sum(self.factors[group].shape[0] for group in self.groups),
                sum(self.weights[view].shape[0] for view in self.views),
            )
        self.nfactors = self.factors[self.groups[0]].shape[0]
        self.nviews = len(self.views)
        self.ngroups = len(self.groups)

        # Load model options
        if "model_options" in self.model:
            self.likelihoods = (
                np.array(self.model["model_options"]["likelihoods"])
                .astype("str")
                .tolist()
            )

        # Load training options
        if "training_opts" in self.model:
            # TODO: Update according to the latest API
            self.training_opts = {"maxiter": self.model["training_opts"][0]}

        # Define samples metadata
        self._samples_metadata = _load_samples_metadata(self)

        # Define features metadata
        self._features_metadata = _load_features_metadata(self)

        ### MEFISTO ###

        # Interpolated Z
        self.interpolated_factors = None

        # Keep the structure similar to self.factors:
        # e.g. self.interpolated_factors["mean"]["group1"]
        # will refer to an HDF5 group with the dataset
        # of shape (n_new_values, n_factors)
        if "Z_predictions" in self.model:
            self.interpolated_factors = dict()
            for attr in "mean", "variance":
                if attr in self.model["Z_predictions"][self.groups[0]]:
                    self.interpolated_factors[attr] = {
                        g: self.model["Z_predictions"][g][attr] for g in self.groups
                    }
            if "new_values" in self.model["Z_predictions"]:
                self.interpolated_factors["new_values"] = self.model["Z_predictions"][
                    "new_values"
                ]

        # Samples covariates
        self.covariates_names, self.covariates = _load_covariates(self)

        # Training stats
        self.training_stats = None
        if "training_stats" in self.model:
            self.training_stats = _read_simple(self.model["training_stats"])

        # Options
        self.options = dict()
        # Training options
        for tr_opts_key in ("training_opts", "training_options"):
            if tr_opts_key in self.model:  # the latter supercedes
                self.options["training"] = _read_simple(self.model[tr_opts_key])
        # Model options
        for mod_opts_key in ("model_opts", "model_options"):
            if mod_opts_key in self.model:
                self.options["model"] = _read_simple(self.model[mod_opts_key])
        # MEFISTO options
        for mod_opts_key in ("smooth_opts", "smooth_options", "mefisto_options"):
            if mod_opts_key in self.model:
                self.options["smooth"] = _read_simple(self.model[mod_opts_key])

    def __repr__(self):
        mofa_repr = f"""MOFA+ model: {" ".join(self.filename.replace(".hdf5", "").split("_"))}
Samples (cells): {self.shape[0]}
Features: {self.shape[1]}
Groups: {', '.join([f"{k} ({len(v)})" for k, v in self.samples.items()])}
Views: {', '.join([f"{k} ({len(v)})" for k, v in self.features.items()])}
Factors: {self.nfactors}
Expectations: {', '.join(self.expectations.keys())}"""

        # MEFISTO
        mefisto_repr = ""
        if self.covariates is not None:
            mefisto_repr += (
                f"\nCovariates available: {', '.join(self.covariates_names)}"
            )
        if self.interpolated_factors is not None:
            mefisto_repr += f"\nInterpolated factors for {str(len(self.interpolated_factors['new_values']))} new values"

        if mefisto_repr != "":
            mofa_repr += "\n\nMEFISTO:" + mefisto_repr

        return mofa_repr

    # Alias samples as cells
    @property
    def cells(self):
        return self.samples

    @property
    def samples_metadata(self):
        return self._samples_metadata

    @samples_metadata.setter
    def samples_metadata(self, metadata):
        if len(metadata) != self.shape[0]:
            raise ValueError(
                f"Length of provided metadata {len(metadata)} does not match the length {self.shape[0]} of the data."
            )
        self._samples_metadata = metadata

    @property
    def cells_metadata(self):
        return self.samples_metadata

    @cells_metadata.setter
    def cells_metadata(self, metadata):
        self.samples_metadata = metadata

    @property
    def metadata(self):
        return self.samples_metadata

    @metadata.setter
    def metadata(self, metadata):
        self.samples_metadata = metadata

    @property
    def features_metadata(self):
        return self._features_metadata

    @features_metadata.setter
    def features_metadata(self, metadata):
        if len(metadata) != self.shape[1]:
            raise ValueError(
                f"Length of provided metadata {len(metadata)} does not match the length {self.shape[1]} of the data."
            )
        self._features_metadata = metadata

    def close(self):
        """Close the connection to the HDF5 file"""
        if self.model.__bool__():  # if the connection is still open
            self.model.close()

    def get_shape(self, groups=None, views=None):
        """
        Get the shape of all the data, samples (cells) and features pulled across groups and views.

        Parameters
        ----------
        groups : optional
            List of groups to consider
        views : optional
            List of views to consider
        """
        groups = self._check_groups(groups)
        views = self._check_views(views)
        shape = (
            sum(self.data[self.views[0]][group].shape[0] for group in groups),
            sum(self.data[view][self.groups[0]].shape[1] for view in views),
        )
        return shape

    def get_samples(self, groups=None):
        """
        Get the sample metadata table (sample ID and its respective group)

        Parameters
        ----------
        groups : optional
            List of groups to consider
        """
        groups = self._check_groups(groups)
        return pd.DataFrame(
            [
                [group, cell]
                for group, cell_list in self.cells.items()
                for cell in cell_list
                if group in groups
            ],
            columns=["group", "sample"],
        )

    # Alias samples as cells
    def get_cells(self, groups=None):
        """
        Get the cell metadata table (cell ID and its respective group)

        Parameters
        ----------
        groups : optional
            List of groups to consider
        """
        cells = self.get_samples(groups)
        cells.columns = ["group", "cell"]
        return cells

    def get_features(self, views=None):
        """
        Get the features metadata table (feature name and its respective view)

        Parameters
        ----------
        views : optional
            List of views to consider
        """
        views = self._check_views(views)
        return pd.DataFrame(
            [
                [view, feature]
                for view, feature_list in self.features.items()
                for feature in feature_list
                if view in views
            ],
            columns=["view", "feature"],
        )

    def get_groups(self):
        """
        Get the groups names
        """
        return self.groups

    def get_views(self):
        """
        Get the views names
        """
        return self.views

    def get_top_features(
        self,
        factors: Union[int, List[int]] = None,
        views: Union[str, int, List[str], List[int]] = None,
        n_features: int = None,
        clip_threshold: float = None,
        scale: bool = False,
        absolute_values: bool = False,
        only_positive: bool = False,
        only_negative: bool = False,
        per_view: bool = True,
        df: bool = False,
    ):
        """
        Fetch a list of top feature names

        Parameters
        ----------
        factors : optional
            Factors to use (all factors in the model by default)
        view : options
            The view to get the factor weights for (first view by default)
        n_features : optional
            Number of features for each factor by their absolute value (10 by default)
        clip_threshold : optional
            Absolute weight threshold to clip all values to (no threshold by default)
        absolute_values : optional
            If to fetch absolute weight values
        only_positive : optional
            If to fetch only positive weights
        only_negative : optional
            If to fetch only negative weights
        per_view : optional
            Get n_features per view rather than globally (True by default)
        df : optional
            Boolean value if to return a DataFrame
        """
        views = self._check_views(views)
        factor_indices, factors = self._check_factors(factors, unique=True)
        n_features_default = 10

        # Fetch weights for the relevant factors
        w = (
            self.get_weights(
                views=views,
                factors=factors,
                df=True,
                absolute_values=absolute_values,
                concatenate_views=True,
            )
            .rename_axis("feature")
            .reset_index()
        )
        wm = w.melt(id_vars="feature", var_name="factor", value_name="value")
        wm = wm.assign(value_abs=lambda x: x.value.abs())
        wm["factor"] = wm["factor"].astype("category")
        wm = (
            wm.set_index("feature")
            .join(self.features_metadata.loc[:, ["view"]], how="left")
            .reset_index()
        )

        if only_positive and only_negative:
            print("Please specify either only_positive or only_negative")
            sys.exit(1)
        elif only_positive:
            wm = wm[wm.value > 0]
        elif only_negative:
            wm = wm[wm.value < 0]

        if n_features is None and clip_threshold is not None:
            wm = wm[wm.value_abs >= clip_threshold]
        else:
            if n_features is None:
                n_features = n_features_default
            # Get a subset of features
            if per_view:
                wm = wm.sort_values(["factor", "value_abs"], ascending=False).groupby(
                    ["factor", "view"]
                )
            else:
                wm = wm.sort_values(["factor", "value_abs"], ascending=False).groupby(
                    ["factor", "view"]
                )
            # Use clip threshold if provided
            if clip_threshold is None:
                wm = wm.head(n_features).reset_index()
            else:
                wm = wm[wm.value_abs >= clip_threshold].head(n_features)

        if df:
            return wm

        features = wm.feature.unique()
        return features

    def get_factors(
        self,
        groups: Union[str, int, List[str], List[int]] = None,
        factors: Optional[Union[int, List[int], str, List[str]]] = None,
        df: bool = False,
        concatenate_groups: bool = True,
        scale: bool = False,
        absolute_values: bool = False,
    ):
        """
        Get the matrix with factors as a NumPy array or as a DataFrame (df=True).

        Parameters
        ----------
        groups : optional
            List of groups to consider
        factors : optional
            Indices of factors to consider
        df : optional
            Boolean value if to return the factor matrix Z as a (wide) pd.DataFrame
        concatenate_groups : optional
            If concatenate Z matrices (True by default)
        scale : optional
            If return values scaled to zero mean and unit variance
            (per factor when concatenated or per factor and per group otherwise)
        absolute_values : optional
            If return absolute values for weights
        """

        groups = self._check_groups(groups)
        factor_indices, factors = self._check_factors(factors)

        # get factors
        z = list(np.array(self.factors[g]).T[:, factor_indices] for g in groups)

        # consider transformations
        for g in range(len(groups)):
            if not concatenate_groups:
                if scale:
                    z[g] = (z[g] - z[g].mean(axis=0)) / z[g].std(axis=0)
                if absolute_values:
                    z[g] = np.absolute(z[g])
            if df:
                z[g] = pd.DataFrame(z[g])
                z[g].columns = factors
                z[g].index = self.samples[groups[g]]

        # concatenate views if requested
        if concatenate_groups:
            z = pd.concat(z) if df else np.concatenate(z)
            if scale:
                z = (z - z.mean(axis=0)) / z.std(axis=0)
            if absolute_values:
                z = np.absolute(z)

        return z

    def get_interpolated_factors(
        self,
        groups: Union[str, int, List[str], List[int]] = None,
        factors: Optional[Union[int, List[int], str, List[str]]] = None,
        df: bool = False,
        df_long: bool = False,
        concatenate_groups: bool = True,
        scale: bool = False,
        absolute_values: bool = False,
    ):
        """
        Get the matrix with interpolated factors.

        If df_long is False, a dictionary with keys ("mean", "variance") is returned
        with NumPy arrays (df=False) or DataFrames (df=True) as values.

        If df_long is True, a DataFrame with columns ("new_value", "factor", "mean", "variance")
        is returned.


        Parameters
        ----------
        groups : optional
            List of groups to consider
        factors : optional
            Indices of factors to consider
        df : optional
            Boolean value if to return mean and variance matrices as (wide) DataFrames
            (can be superseded by df_long=True)
        df_long : optional
            Boolean value if to return a single long DataFrame
            (supersedes df=False and concatenate_groups=False)
        concatenate_groups : optional
            If concatenate Z matrices (True by default, can be superseded by df_long=True)
        scale : optional
            If return values scaled to zero mean and unit variance
            (per factor when concatenated or per factor and per group otherwise)
        absolute_values : optional
            If return absolute values for weights
        """

        groups = self._check_groups(groups)
        factor_indices, factors = self._check_factors(factors)

        z_interpolated = dict()
        new_values_names = tuple()
        if self.covariates_names:
            new_values_names = tuple(
                [f"{value}_transformed" for value in self.covariates_names]
            )
        else:
            new_values_names = tuple(
                [
                    f"new_value{i}"
                    for i in range(self.interpolated_factors["new_values"].shape[1])
                ]
            )

        for stat in ["mean", "variance"]:
            # get factors
            z = list(
                np.array(self.interpolated_factors[stat][g])[:, factor_indices]
                for g in groups
            )

            # consider transformations
            for g in range(len(groups)):
                if not concatenate_groups:
                    if scale:
                        z[g] = (z[g] - z[g].mean(axis=0)) / z[g].std(axis=0)
                    if absolute_values:
                        z[g] = np.absolute(z[g])
                if df or df_long:
                    z[g] = pd.DataFrame(z[g])
                    z[g].columns = factors
                    z[g]["group"] = self.groups[g]

                    if "new_values" in self.interpolated_factors:
                        new_values = np.array(self.interpolated_factors["new_values"])
                    else:
                        new_values = np.arange(z[g].shape[0]).reshape(-1, 1)

                    new_values = pd.DataFrame(new_values, columns=new_values_names)

                    z[g] = pd.concat([z[g], new_values], axis=1)

                    # If groups are to be concatenated (but not in a long DataFrame),
                    # index has to be made unique per group
                    new_samples = [
                        f"{groups[g]}_{'_'.join(value.astype(str))}"
                        for _, value in new_values.iterrows()
                    ]

                    z[g].index = new_samples

                    # Create an index for new values
                    z[g]["new_value"] = np.arange(z[g].shape[0])

            # concatenate views if requested
            if concatenate_groups:
                z = pd.concat(z) if df or df_long else np.concatenate(z)
                if scale:
                    z = (z - z.mean(axis=0)) / z.std(axis=0)
                if absolute_values:
                    z = np.absolute(z)

            # melt DataFrames
            if df_long:
                if not concatenate_groups:  # supersede
                    z = pd.concat(z)
                z = (
                    z.rename_axis("new_sample", axis=0)
                    .reset_index()
                    .melt(
                        id_vars=["new_sample", *new_values_names, "group"],
                        var_name="factor",
                        value_name=stat,
                    )
                )

            z_interpolated[stat] = z

        if df_long:
            z_interpolated = (
                z_interpolated["mean"]
                .set_index(["new_sample", *new_values_names, "group", "factor"])
                .merge(
                    z_interpolated["variance"],
                    on=("new_sample", *new_values_names, "group", "factor"),
                )
            )

        return z_interpolated

    def get_group_kernel(self):
        model_groups = False
        if (
            self.options
            and "smooth" in self.options
            and "model_groups" in self.options["smooth"]
        ):
            model_groups = bool(self.options["smooth"]["model_groups"].item().decode())

        kernels = list()
        if not model_groups or self.ngroups == 1:
            Kg = np.ones(shape=(self.nfactors, self.ngroups, self.ngroups))
            return Kg
        else:
            if self.training_stats and "Kg" in self.training_stats:
                return self.training_stats["Kg"]
            else:
                raise ValueError(
                    "No group kernel was saved. Specify the covariates and train the MEFISTO model with the option 'model_groups' set to True."
                )

    def get_weights(
        self,
        views: Union[str, int, List[str], List[int]] = None,
        factors: Union[int, List[int]] = None,
        df: bool = False,
        scale: bool = False,
        concatenate_views: bool = True,
        absolute_values: bool = False,
    ):
        """
        Fetch the weight matrices

        Parameters
        ----------
        views : optional
            List of views to consider
        factors : optional
            Indices of factors to use
        df : optional
            Boolean value if to return W matrix as a (wide) pd.DataFrame
        scale : optional
            If return values scaled to zero mean and unit variance
            (per factor when concatenated or per factor and per view otherwise)
        concatenate_weights : optional
            If concatenate W matrices (True by default)
        absolute_values : optional
            If return absolute values for weights
        """

        views = self._check_views(views)
        factor_indices, factors = self._check_factors(factors, unique=True)

        # get views
        w = list(np.array(self.weights[m]).T[:, factor_indices] for m in views)

        # consider transformations
        for m in range(len(views)):
            if not concatenate_views:
                if scale:
                    w[m] = (w[m] - w[m].mean(axis=0)) / w[m].std(axis=0)
                if absolute_values:
                    w[m] = np.absolute(w[m])
            if df:
                w[m] = pd.DataFrame(w[m])
                w[m].columns = factors
                w[m].index = self.features[views[m]]

        # concatenate views if requested
        if concatenate_views:
            w = pd.concat(w) if df else np.concatenate(w)
            if scale:
                w = (w - w.mean(axis=0)) / w.std(axis=0)
            if absolute_values:
                w = np.absolute(w)

        return w

    def get_data(
        self,
        views: Optional[Union[str, int]] = None,
        features: Optional[Union[str, List[str]]] = None,
        groups: Optional[Union[str, int, List[str], List[int]]] = None,
        df: bool = False,
    ):
        """
        Fetch the training data

        Parameters
        ----------
        view : optional
            view to consider
        features : optional
            Features to consider (from one view)
        groups : optional
            groups to consider
        df : optional
            Boolean value if to return Y matrix as a DataFrame
        """

        # Sanity checks
        groups = self._check_groups(groups)
        views = self._check_views(views)

        # If features is None (default), return all by default
        pd_features = self.get_features(views)
        if features is None:
            features = pd_features.feature.values

        # If a sole feature name is used, wrap it in a list
        if not isinstance(features, Iterable) or isinstance(features, str):
            features = [features]
        else:
            features = list(set(features))  # make feature names unique

        f_i = np.where(pd_features.feature.isin(features))[0]
        assert len(f_i) > 0, "Requested features are not found"
        pd_features = pd_features.loc[f_i]

        # Create numpy array
        # y = [self.data[view][g][:, :] for g in groups]
        ym = []
        for m in views:
            ym.append(np.concatenate([self.data[m][g][:, f_i] for g in groups], axis=0))
        y = np.concatenate(ym, axis=1)

        # Convert output to pandas data.frame
        if df:
            y = pd.DataFrame(y)
            y.columns = pd_features.feature.values
            y.index = np.concatenate(tuple(self.samples[g] for g in groups))

        return y

    def run_umap(
        self,
        groups: Union[str, int, List[str], List[int]] = None,
        factors: Union[int, List[int]] = None,
        n_neighbors: int = 10,
        min_dist: float = 0.5,
        spread: float = 1.0,
        random_state: int = 42,
        **kwargs,
    ) -> None:
        """
        Run UMAP on the factor space

        Parameters
        ----------
        n_neighbors : optional
            UMAP parameter: number of neighbors.
        min_dist
            UMAP parameter: the effective minimum distance between embedded points. Smaller values
            will result in a more clustered/clumped embedding where nearby points on
            the manifold are drawn closer together, while larger values will result
            on a more even dispersal of points. The value should be set relative to
            the ``spread`` value, which determines the scale at which embedded
            points will be spread out.
        spread
            UMAP parameter: the effective scale of embedded points. In combination with `min_dist`
            this determines how clustered/clumped the embedded points are.
        random_state
            random seed
        """
        import umap

        # Get factors
        data = self.get_factors(groups, factors)

        embedding = umap.UMAP(
            n_neighbors=n_neighbors,
            min_dist=min_dist,
            spread=spread,
            random_state=random_state,
            **kwargs,
        ).fit_transform(data)

        # create pandas dataframe
        pd_umap = pd.DataFrame(embedding)
        pd_umap.columns = ["UMAP" + str(i + 1) for i in range(embedding.shape[1])]
        pd_umap.index = self.get_samples()["sample"]

        # merge with samples metadata
        self.samples_metadata = pd.merge(
            left=self._samples_metadata,
            right=pd_umap,
            how="left",
            left_on="sample",
            right_on="sample",
        )
        print("UMAP coordinates added to the samples_metadata")

    def fetch_values(self, variables: Union[str, List[str]], unique: bool = True):
        """
        Fetch metadata column, factors, or feature values
        as well as covariates.
        Shorthand to get_data, get_factors, metadata, and covariates calls.

        Parameters
        ----------
        variables : str
            Features, metadata columns, or factors (FactorN) to fetch.
            For MEFISTO models with covariates, covariates are accepted
            such as 'covariate1' or 'covariate1_transformed'.
        """
        # If a sole variable name is used, wrap it in a list
        if not isinstance(variables, Iterable) or isinstance(variables, str):
            variables = [variables]

        # Remove None values and duplicates
        variables = [i for i in variables if i is not None]
        # Transform integers to factors
        variables = maybe_factor_indices_to_factors(variables)
        if unique:
            variables = pd.Series(variables).drop_duplicates().tolist()

        var_meta = list()
        var_features = list()
        var_factors = list()
        var_covariates = list()

        # Split all the variables into metadata and features
        for i, var in enumerate(variables):
            if var in self.metadata.columns:
                var_meta.append(var)
            elif var.capitalize().startswith("Factor"):
                # Unify factor naming
                variables[i] = var.capitalize()
                var_factors.append(var.capitalize())
            elif (
                self.covariates_names is not None
                and (
                    var in self.covariates_names
                    or var in [f"{cov}_transformed" for cov in self.covariates_names]
                )
                and self.covariates is not None
            ):
                var_covariates.append(var)
            else:
                var_features.append(var)

        var_list = list()
        if len(var_meta) > 0:
            var_list.append(self.metadata[var_meta])
        if len(var_features) > 0:
            var_list.append(self.get_data(features=var_features, df=True))
        if len(var_factors) > 0:
            var_list.append(self.get_factors(factors=var_factors, df=True))
        if len(var_covariates) > 0:
            var_list.append(self.covariates[var_covariates])

        # Return a DataFrame with columns ordered as requested
        return pd.concat(var_list, axis=1).loc[:, variables]

    def _check_views(self, views):
        if views is None:
            views = self.views
        # single view provided as a string
        elif isinstance(views, str):
            views = [views]

        # single view provided as an integer
        elif isinstance(views, int):
            views = [self.views[views]]

        # multiple views provided as an iterable
        elif isinstance(views, Iterable) and not isinstance(views, str):

            # (to-do) check that all elements are of the same type

            # iterable of booleans
            if all([isinstance(m, bool) for m in views]):
                raise ValueError(
                    f"Please provide view names as string or view indices as integers, boolean values are not accepted. Group names of this model are {', '.join(self.views)}."
                )
            # iterable of integers
            elif all([isinstance(m, int) for m in views]):
                views = [self.views[m] if isinstance(m, int) else m for m in views]
            # iterable of strings
            elif all([isinstance(m, str) for m in views]):
                assert set(views).issubset(
                    set(self.views)
                ), f"some of the elements of the 'views' are not valid views. Views names of this model are {', '.join(self.views)}."
            else:
                raise ValueError(
                    "elements of the 'view' vector have to be either integers or strings"
                )
        else:
            raise ValueError("views argument not recognised")

        return views

    def _check_groups(self, groups):
        if groups is None:
            groups = self.groups
        # single group provided as a string
        elif isinstance(groups, str):
            groups = [groups]

        # single group provided as an integer
        elif isinstance(groups, int):
            groups = [self.groups[groups]]

        # multiple groups provided as an iterable
        elif isinstance(groups, Iterable) and not isinstance(groups, str):

            # (to-do) check that all elements are of the same type

            # iterable of booleans
            if all([isinstance(g, bool) for g in groups]):
                raise ValueError(
                    f"Please provide group names as string or group indices as integers, boolean values are not accepted. Group names of this model are {', '.join(self.groups)}."
                )
            # iterable of integers
            elif all([isinstance(g, int) for g in groups]):
                groups = [self.groups[g] if isinstance(g, int) else g for g in groups]
            # iterable of strings
            elif all([isinstance(g, str) for g in groups]):
                assert set(groups).issubset(
                    set(self.groups)
                ), f"some of the elements of the 'groups' are not valid groups. Group names of this model are {', '.join(self.groups)}."
            else:
                raise ValueError(
                    "elements of the 'group' vector have to be either integers or strings"
                )
        else:
            raise ValueError("groups argument not recognised")

        return groups

    # def _check_grouping(self, groups, grouping_instance):
    #     assert grouping_instance in ["groups", "views"]
    #     # Use all groups if no specific groups are requested
    #     if groups is None:
    #         if grouping_instance == "groups":
    #             groups = self.groups
    #         elif grouping_instance == "views":
    #             groups = self.views
    #     # If a sole group name is used, wrap it in a list
    #     if not isinstance(groups, Iterable) or isinstance(groups, str):
    #         groups = [groups]
    #     # Do not accept boolean values
    #     if any([isinstance(g, bool) for g in groups]):
    #         if grouping_instance == "groups":
    #             raise ValueError(
    #                 f"Please provide relevant group names. Boolean values are not accepted. Group names of this model are {', '.join(self.groups)}."
    #             )
    #         elif grouping_instance == "views":
    #             raise ValueError(
    #                 f"Please provide relevant view names. Boolean values are not accepted. View names of this model are {', '.join(self.views)}."
    #             )
    #     # Convert integers to group names
    #     if grouping_instance == "groups":
    #         groups = [self.groups[g] if isinstance(g, int) else g for g in groups]
    #     elif grouping_instance == "views":
    #         groups = [self.views[g] if isinstance(g, int) else g for g in groups]
    #     return groups

    def _check_factors(self, factors, unique=False):
        # Use all factors by default
        if factors is None:
            factors = list(range(self.nfactors))
        # If one factor is used, wrap it in a list
        if not isinstance(factors, Iterable) or isinstance(factors, str):
            factors = [factors]
        if unique:
            factors = list(set(factors))
        # Convert factor names (FactorN) to factor indices (N-1)
        factor_indices = [
            int(fi.replace("Factor", "")) - 1 if isinstance(fi, str) else fi
            for fi in factors
        ]
        factors = [f"Factor{fi+1}" if isinstance(fi, int) else fi for fi in factors]

        return (factor_indices, factors)

    # Variance explained (R2)

    def calculate_variance_explained(
        self,
        # factor_index: int,
        factors: Optional[Union[int, List[int], str, List[str]]] = None,
        groups: Optional[Union[str, int, List[str], List[int]]] = None,
        views: Optional[Union[str, int, List[str], List[int]]] = None,
        group_label: Optional[str] = None,
    ) -> pd.DataFrame:
        """
        Calculate the variance explained estimates for each factor in each view and/or group.
        Allow also for predefined groups

        Parameters
        ----------
        factors : optional
            List of factors to consider (default is None, all factors)
        groups : optional
            List of groups to consider (default is None, all groups)
        views : optional
            List of views to consider (default is None, all views)
        """

        groups = self._check_groups(groups)
        views = self._check_views(views)
        factor_indices, _ = self._check_factors(factors)

        r2_df = pd.DataFrame()

        # use model groups
        if group_label is None or group_label == "group":
            for view in views:
                for group in groups:
                    r2 = calculate_r2(
                        Z=np.array(self.expectations["Z"][group][factor_indices, :]),
                        W=np.array(self.expectations["W"][view][factor_indices, :]),
                        Y=np.array(self.data[view][group]),
                    )
                    r2_df = r2_df.append(
                        {"View": view, "Group": group, "R2": r2}, ignore_index=True
                    )

        # use custom groups
        # note that when calculating for a custom set of groups,
        # the Factor matrix (Z) has to be merged and then split according to the new grouping of samples
        else:
            custom_groups = self.samples_metadata[group_label].unique()
            samples_groups = self.samples_metadata[group_label]

            z = np.concatenate(
                [self.expectations["Z"][group][:, :] for group in groups], axis=1
            )

            z_custom = dict()
            for group in custom_groups:
                z_custom[group] = z[:, np.where(samples_groups == group)[0]]
            del z

            for view in views:

                y_view = np.concatenate(
                    [self.data[view][group][:, :] for group in groups], axis=0
                )

                data_view = dict()
                for group in custom_groups:
                    data_view[group] = y_view[np.where(samples_groups == group)[0], :]

                for group in custom_groups:
                    r2 = calculate_r2(
                        Z=np.array(z_custom[group][factor_indices, :]),
                        W=np.array(self.expectations["W"][view][factor_indices, :]),
                        Y=np.array(data_view[group]),
                    )
                    r2_df = r2_df.append(
                        {"View": view, "Group": group, "R2": r2}, ignore_index=True
                    )
        return r2_df

    def get_variance_explained(
        self,
        factors: Optional[Union[int, List[int], str, List[str]]] = None,
        groups: Optional[Union[str, int, List[str], List[int]]] = None,
        views: Optional[Union[str, int, List[str], List[int]]] = None,
    ) -> pd.DataFrame:
        """
        Get variance explained estimates (R2) for each factor across  view(s) and/or group(s).

        factors : optional
            List of factors to consider (all by default)
        groups : optional
            List of groups to consider (all by default)
        views : optional
            List of views to consider (all by default)
        """

        # Load from file if pre-computed
        if "variance_explained" in self.model.keys():
            r2 = pd.concat(
                [
                    pd.DataFrame(
                        r2,
                        index=self.views,
                        columns=[f"Factor{i+1}" for i in range(self.nfactors)],
                    )
                    .rename_axis("View")
                    .reset_index()
                    .melt(id_vars=["View"], var_name="Factor", value_name="R2")
                    .assign(Group=group)
                    .loc[:, ["Factor", "View", "Group", "R2"]]
                    for group, r2 in self.model["variance_explained"][
                        "r2_per_factor"
                    ].items()
                ]
            )
            # Choose factors of interest
            factor_indices, factors = self._check_factors(factors, unique=True)
            r2 = r2[r2.Factor.isin(factors)]
        # Recalculate if not pre-computed
        else:

            r2 = pd.DataFrame()
            factor_indices, _ = self._check_factors(factors)
            for k in factor_indices:
                tmp = self.calculate_variance_explained(
                    factors=k, groups=groups, views=views
                )
                tmp["Factor"] = "Factor" + str(k)
                r2 = r2.append(tmp)

            # Subset
            if factors is not None:
                _, factors = self._check_factors(factors)
                r2 = r2[r2.Factor.isin(factors)]
            if groups is not None:
                groups = self._check_groups(groups)
                r2 = r2[r2.Group.isin(groups)]
            if views is not None:
                view = self._check_views(views)
                r2 = r2[r2.View.isin(views)]
        return r2

    def get_r2(
        self,
        factors: Optional[Union[int, List[int], str, List[str]]] = None,
        groups: Optional[Union[str, int, List[str], List[int]]] = None,
        views: Optional[Union[str, int, List[str], List[int]]] = None,
        groups_df: Optional[pd.DataFrame] = None,
        group_label: Optional[str] = None,
    ) -> pd.DataFrame:
        """
        Get variance explained (R2) per factor, view, and group.
        factors : optional
            List of factors to consider (all by default)
        groups : optional
            List of groups to consider (all by default)
        views : optional
            List of views to consider (all by default)
        group_label : optional
            Sample (cell) metadata column to be used as group assignment
        groups_df : optional pd.DataFrame
            Data frame with samples (cells) as index and first column as group assignment
        """
        warnings.warn(
            "This method will be deprecated. Please use `.get_variance_explained`",
            DeprecationWarning,
        )
        if groups_df is not None:
            raise ValueError(
                "Argument groups_df is not supported anymore. Please use `group_label` and sample metadata instead"
            )
        if group_label is None:
            return self.get_variance_explained(
                factors=factors, groups=groups, views=views
            )
        else:
            return self.calculate_variance_explained(
                factors=factors, groups=groups, views=views, group_label=group_label
            )

    def _get_factor_r2_null(
        self,
        factor_index: int,
        groups_df: Optional[pd.DataFrame],
        group_label: Optional[str],
        n_iter=100,
        return_full=False,
        return_true=False,
        return_pvalues=True,
        fdr=True,
    ) -> pd.DataFrame:
        r2_df = pd.DataFrame()

        if groups_df is None and group_label is None:
            group_label = "group"

        if groups_df is None:
            groups_df = self.samples_metadata.loc[:, [group_label]]

        custom_groups = groups_df.iloc[:, 0].unique()

        z = np.concatenate(
            [self.expectations["Z"][group][:, :] for group in self.groups], axis=1
        )

        for i in range(n_iter + 1):
            # Calculate true group assignment for iteration 0
            if i > 0:
                groups_df.iloc[:, 0] = groups_df.iloc[:, 0].sample(frac=1).values

            z_custom = dict()
            for group in custom_groups:
                z_custom[group] = z[:, np.where(groups_df.iloc[:, 0] == group)[0]]

            for view in self.views:

                y_view = np.concatenate(
                    [self.data[view][group][:, :] for group in self.groups], axis=0
                )

                data_view = dict()
                for group in custom_groups:
                    data_view[group] = y_view[
                        np.where(groups_df.iloc[:, 0] == group)[0], :
                    ]

                for group in custom_groups:
                    crossprod = np.array(z_custom[group][[factor_index], :]).T.dot(
                        np.array(self.expectations["W"][view][[factor_index], :])
                    )
                    y = np.array(data_view[group])
                    a = np.sum((y - crossprod) ** 2)
                    b = np.sum(y ** 2)
                    r2_df = r2_df.append(
                        {
                            "View": view,
                            "Group": group,
                            "Factor": f"Factor{factor_index+1}",
                            "R2": 1 - a / b,
                            "Iteration": i,
                        },
                        ignore_index=True,
                    )

        if return_full:
            if return_true:
                return r2_df
            else:
                return r2_df[r2_df.Iteration != 0].reset_index(drop=True)

        r2_obs = r2_df[r2_df.Iteration == 0]
        r2_df = r2_df[r2_df.Iteration != 0]

        if not return_pvalues:
            r2_null = r2_df.groupby(["Factor", "Group", "View"]).agg(
                {"R2": ["mean", "std"]}
            )
            return r2_null.reset_index()

        r2_pvalues = pd.DataFrame(
            r2_obs.set_index(["Group", "View", "Factor"])
            .loc[:, ["R2"]]
            .join(r2_df.set_index(["Group", "View", "Factor"]), rsuffix="_null")
            .groupby(["Group", "View", "Factor"])
            .apply(lambda x: np.mean(x["R2"] <= x["R2_null"]))
        )
        r2_pvalues.columns = ["PValue"]

        if fdr:
            r2_pvalues["FDR"] = padjust_fdr(r2_pvalues.PValue)
            return r2_pvalues.reset_index().sort_values("FDR", ascending=True)
        else:
            return r2_pvalues.reset_index().sort_values("PValue", ascending=True)

    def _get_r2_null(
        self,
        factors: Union[int, List[int], str, List[str]] = None,
        n_iter: int = 100,
        groups_df: Optional[pd.DataFrame] = None,
        group_label: Optional[str] = None,
        return_full=False,
        return_pvalues=True,
        fdr=True,
    ) -> pd.DataFrame:
        factor_indices, factors = self._check_factors(factors)
        r2 = pd.DataFrame()
        for fi in factor_indices:
            r2 = r2.append(
                self._get_factor_r2_null(
                    fi,
                    groups_df=groups_df,
                    group_label=group_label,
                    n_iter=n_iter,
                    return_full=return_full,
                    return_pvalues=return_pvalues,
                    fdr=fdr,
                )
            )
        return r2

    def get_sample_r2(
        self,
        factors: Optional[Union[str, int, List[str], List[int]]] = None,
        groups: Optional[Union[str, int, List[str], List[int]]] = None,
        views: Optional[Union[str, int, List[str], List[int]]] = None,
        df: bool = True,
    ) -> pd.DataFrame:
        findices, factors = self.__check_factors(factors, unique=True)
        groups = self.__check_groups(groups)
        views = self.__check_views(views)

        r2s = []
        for view in views:
            for group in groups:
                crossprod = self.expectations["Z"][group][findices, :].T.dot(
                    self.expectations["W"][view][findices, :]
                )
                y = np.array(self.data[view][group])
                a = np.nansum((y - crossprod) ** 2.0, axis=1)
                b = np.nansum(y ** 2, axis=1)

                r2_df_mg = pd.DataFrame(
                    {
                        "Sample": self.samples[group],
                        "Group": group,
                        "View": view,
                        "R2": (1.0 - a / b),
                    }
                )

                r2s.append(r2_df_mg)

        r2_df = pd.concat(r2s, axis=0, ignore_index=True)

        if df:
            return r2_df
        else:
            r2_mx = np.array(r2_df.pivot(index="Sample", columns="View", values="R2"))
            return r2_mx

    def project_data(
        self,
        data,
        view: Union[str, int] = None,
        factors: Union[int, List[int], str, List[str]] = None,
        df: bool = False,
        feature_intersection: bool = False,
    ):
        """
        Project new data onto the factor space of the model.

        For the projection, a pseudo-inverse of the weights matrix is calculated
        and its product with the provided data matrix is calculated.

        Parameters
        ----------
        data
            Numpy array or Pandas DataFrame with the data matching the number of features
        view : optional
            A view of the model to consider (first view by default)
        factors : optional
            Indices of factors to use for the projection (all factors by default)
        """
        if view is None:
            view = 0
        view = self._check_views([view])[0]
        factor_indices, factors = self._check_factors(factors)

        # Calculate the inverse of W
        winv = np.linalg.pinv(self.get_weights(views=view, factors=factors))
        winv = np.concatenate(winv)

        # Find feature intersection to match the dimensions
        if feature_intersection:
            if data.shape[1] != self.shape[1] and isinstance(data, pd.DataFrame):
                fs_common = np.intersect1d(data.columns.values, self.features[view])
                data = data.loc[:, fs_common]

                # Get indices of the common features in the original data
                f_sorted = np.argsort(self.features[view])
                fs_common_pos = np.searchsorted(
                    self.features[view][f_sorted], fs_common
                )
                f_indices = f_sorted[fs_common_pos]

                winv = winv[:, f_indices]
                warnings.warn(
                    "Only {} features are matching between two datasets of size {} (original data) and {} (projected data).".format(
                        fs_common.shape[0], self.shape[1], data.shape[1]
                    )
                )

        # Predict Z for the provided data
        zpred = np.dot(data, winv.T)

        if df:
            zpred = pd.DataFrame(zpred)
            zpred.columns = factors
            if isinstance(data, pd.DataFrame):
                zpred.index = data.index
        return zpred

    def get_views_contributions(self, scaled: bool = True):
        """
        Project new data onto the factor space of the model.

        For the projection, a pseudo-inverse of the weights matrix is calculated
        and its product with the provided data matrix is calculated.

        Parameters
        ----------
        scaled : bool, optional
            Whether to scale contributions scores per sample
            so that they sum up to 1 (True by default)

        Returns
        -------
        pd.DataFrame
            Dataframe with view contribution scores, samples in rows and views in columns
        """
        z = self.get_factors()
        z = np.abs(z) / np.max(np.abs(z), axis=0)

        factors_ordered = self.get_r2().Factor.drop_duplicates().values

        contributions = []
        r2 = self.get_r2()
        for g in self.groups:
            r2_g = r2[r2.apply(lambda x: x.Group == g, axis=1)]
            z_g_indices = self.metadata.group.values == g
            z_g = z[z_g_indices]
            # R2 per factor
            r2_per_factor = r2_g.pivot(index="Factor", columns="View", values="R2").loc[
                factors_ordered, self.views
            ]
            # R2 per view
            r2_per_view = np.array(self.model["variance_explained"]["r2_total"][g])
            # Z x R2 per factor
            view_contribution = np.dot(z_g, r2_per_factor) / r2_per_view
            if scaled:
                # Scale contributions to sum to 1
                view_contribution = (
                    view_contribution / view_contribution.sum(axis=1)[:, None]
                )
            view_contribution = pd.DataFrame(
                view_contribution,
                index=self.metadata.index[z_g_indices],
                columns=r2_per_factor.columns,
            )
            contributions.append(view_contribution)

        view_contribution = pd.concat(contributions, axis=0)

        return view_contribution



from .core import mofa_model
from .utils import *

import sys
from warnings import warn
from typing import Union, Optional, List, Iterable, Sequence
from functools import partial

import numpy as np
from scipy.stats import pearsonr
import pandas as pd
from pandas.api.types import is_numeric_dtype
import matplotlib.pyplot as plt
from matplotlib import rcParams
import seaborn as sns

from .utils import maybe_factor_indices_to_factors, _make_iterable, _is_iter
from .plot_utils import _plot_grid


### WEIGHTS ###


def plot_weights(
    model: mofa_model,
    factors=None,
    views=None,
    n_features: int = 5,
    w_scaled: bool = False,
    w_abs: bool = False,
    size=2,
    color="black",
    label_size=5,
    x_offset=0.01,
    y_offset=0.15,
    jitter=0.01,
    line_width=0.5,
    line_color="black",
    line_alpha=0.2,
    zero_line=True,
    zero_line_width=1,
    ncols=4,
    sharex=True,
    sharey=False,
    **kwargs,
):
    """
    Plot weights for a specific factor

    Parameters
    ----------
    model : mofa_model
        An instance of the mofa_model class
    factors : str or int or list of str or None
        Factors to use (default is all)
    views : str or int or list of str or None
        The views to get the factors weights for (first view by default)
    n_features : int
        Number of features with the largest weights to label (in absolute values)
    w_scaled : bool
        If scale weights to unite variance (False by default)
    w_abs : bool
        If plot absolute weight values (False by default)
    size : float
        Dot size (2 by default)
    color : str
        Colour for the labelled dots (black by default)
    label_size : int or float
        Font size of feature labels (default is 5)
    x_offset : int or float
        Offset the feature labels from the left/right side (by 0.03 points by default)
    y_offset : int or float
        Parameter to repel feature labels along the y axis (0.1 by default)
    jitter : bool
        Jitter dots per factors (True by default)
    line_width : int or float
        Width of the lines connecting labels with dots (0.5 by default)
    line_color : str
        Color of the lines connecting labels with dots (black by default)
    line_alpha : float
        Alpha level for the lines connecting labels with dots (0.2 by default)
    zero_line : bool
        If to plot a dotted line at zero (False by default)
    zero_line_width : int or float
        Width of the line at 0 (1 by default)
    ncols : int
        Number of columns in the grid of multiple plots, one plot per view (4 by default)
    sharex : bool
        If to use the same X axis across panels (True by default)
    sharey : bool
        If to use the same Y axis across panels (False by default)
    """
    w = model.get_weights(
        views=views,
        factors=factors,
        df=True,
        scale=w_scaled,
        absolute_values=w_abs,
    )
    wm = (
        w.join(model.features_metadata.loc[:, ["view"]])
        .rename_axis("feature")
        .reset_index()
        .melt(id_vars=["feature", "view"], var_name="factor", value_name="value")
    )

    wm["abs_value"] = abs(wm.value)

    # Assign ranks to features, per factor
    wm["rank"] = wm.groupby("factor")["value"].rank(ascending=False)
    wm["abs_rank"] = wm.groupby("factor")["abs_value"].rank(ascending=False)
    wm = wm.sort_values(["factor", "abs_rank"], ascending=True)

    # Sort factors
    wm["factor"] = wm["factor"].astype("category")
    wm["factor"] = wm["factor"].cat.reorder_categories(
        sorted(wm["factor"].cat.categories, key=lambda x: int(x.split("Factor")[1]))
    )

    # Set default colour to black if none set
    if "c" not in kwargs and "color" not in kwargs:
        kwargs["color"] = "black"

    # Fetch top features to label
    features_to_label = model.get_top_features(
        factors=factors, views=views, n_features=n_features, df=True
    )
    features_to_label["to_label"] = True
    wm = (
        features_to_label.loc[:, ["feature", "view", "factor", "to_label"]]
        .set_index(["feature", "view", "factor"])
        .join(wm.set_index(["feature", "factor", "view"]), how="right")
        .reset_index()
        .fillna({"to_label": False})
        .sort_values(["factor", "to_label"])
    )

    # Figure out rows & columns for the grid with plots (one plot per view)
    view_vars = wm.view.unique()
    ncols = min(ncols, len(view_vars))
    nrows = int(np.ceil(len(view_vars) / ncols))
    fig, axes = plt.subplots(
        nrows,
        ncols,
        sharex=sharex,
        sharey=sharey,
        figsize=(
            ncols * rcParams["figure.figsize"][0],
            nrows * rcParams["figure.figsize"][1],
        ),
    )
    if ncols == 1:
        axes = np.array(axes).reshape(-1, 1)
    if nrows == 1:
        axes = np.array(axes).reshape(1, -1)

    for m, view in enumerate(view_vars):
        ri = m // ncols
        ci = m % ncols

        wm_view = wm.query("view == @view")

        # Construct the plot
        g = sns.stripplot(
            data=wm_view,
            x="value",
            y="factor",
            jitter=jitter,
            size=size,
            hue="to_label",
            palette=["lightgrey", color],
            ax=axes[ri, ci],
        )
        sns.despine(offset=10, trim=True, ax=g)
        g.legend().remove()

        # Label some points
        for fi, factor in enumerate(wm_view.factor.cat.categories):
            for sign_i in [1, -1]:
                to_label = features_to_label.query(
                    "factor == @factor & view == @view"
                ).feature.tolist()
                w_set = wm_view.query(
                    "factor == @factor & value * @sign_i > 0 & feature == @to_label & view == @view"
                ).sort_values("abs_value", ascending=False)

                x_start_pos = sign_i * (w_set.abs_value.max() + x_offset)
                y_start_pos = fi - ((w_set.shape[0] - 1) // 2) * y_offset
                y_prev = y_start_pos

                for i, row in enumerate(w_set.iterrows()):
                    name, point = row
                    y_loc = y_prev + y_offset if i != 0 else y_start_pos

                    g.annotate(
                        point["feature"],
                        xy=(point.value, fi),
                        xytext=(x_start_pos, y_loc),
                        arrowprops=dict(
                            arrowstyle="-",
                            connectionstyle="arc3",
                            color=line_color,
                            alpha=line_alpha,
                            linewidth=line_width,
                        ),
                        horizontalalignment="left" if sign_i > 0 else "right",
                        size=label_size,
                        color="black",
                        weight="regular",
                        alpha=0.9,
                    )
                    y_prev = y_loc

        # Set plot axes labels
        g.set(ylabel="", xlabel="Feature weight", title=view)

        if zero_line:
            axes[ri, ci].axvline(
                0, ls="--", color="lightgrey", linewidth=zero_line_width, zorder=0
            )

    # Remove unused axes
    for i in range(len(view_vars), ncols * nrows):
        ri = i // ncols
        ci = i % ncols
        fig.delaxes(axes[ri, ci])

    return g


def plot_weights_ranked(
    model: mofa_model,
    factor="Factor1",
    view=0,
    n_features: int = 10,
    size: int = 25,
    label_size=5,
    x_rank_offset=10,
    x_rank_offset_neg=0,
    y_repel_coef=0.03,
    attract_to_points=True,
    **kwargs,
):
    """
    Plot weights for a specific factor

    Parameters
    ----------
    model : mofa_model
        Factor model
    factor : optional
        Factor to use (default is Factor1)
    view : options
        The view to get the factors weights for (first view by default)
    n_features : optional
        Number of features to label with most positive and most negative weights
    size : int
        Dit size for labelled features (default is 25)
    label_size : optional
        Font size of feature labels (default is 5)
    x_rank_offset : optional
        Offset the feature labels from the left/right side (by 10 points by default)
    x_rank_offset_neg : optional
        Offset but for the negative weights only (i.e. from the right side)
    y_repel_coef : optional
        Parameter to repel feature labels along the y axis (0.03 by default)
    attract_to_points : optional
        If place labels according to the Y coordinate of the point (False by default)
    """
    w = model.get_weights(views=view, factors=factor, df=True)
    w = pd.melt(
        w.reset_index().rename(columns={"index": "feature"}),
        id_vars="feature",
        var_name="factor",
        value_name="value",
    )
    w["abs_value"] = abs(w.value)

    # Assign ranks to features, per factor
    w["rank"] = w.groupby("factor")["value"].rank(ascending=False)
    w["abs_rank"] = w.groupby("factor")["abs_value"].rank(ascending=False)
    w = w.sort_values(["factor", "abs_rank"], ascending=True)

    # Set default colour to black if none set
    if "c" not in kwargs and "color" not in kwargs:
        kwargs["color"] = "black"

    # Construct the plot
    ax = sns.lineplot(
        x="rank", y="value", data=w, markers=True, dashes=False, linewidth=0.5, **kwargs
    )
    sns.despine(offset=10, trim=True, ax=ax)

    # Plot top features as dots
    sns.scatterplot(
        x="rank",
        y="value",
        data=w[w["abs_rank"] < n_features],
        linewidth=0.2,
        s=size,
        alpha=0.75,
        **kwargs,
    )

    # Label top features

    # Positive weights
    y_start_pos = w[w.value > 0].sort_values("abs_rank").iloc[0].value

    y_prev = y_start_pos
    for i, point in (
        w[(w["abs_rank"] < n_features) & (w["value"] >= 0)].reset_index().iterrows()
    ):
        y_loc = y_prev - y_repel_coef if i != 0 else y_start_pos
        y_loc = min(point["value"], y_loc) if attract_to_points else y_loc
        ax.text(
            x_rank_offset,
            y_loc,
            point["feature"],
            horizontalalignment="left",
            size=label_size,
            color="black",
            weight="regular",
        )
        y_prev = y_loc

    # Negative weights
    y_start_neg = w[w.value < 0].sort_values("abs_rank").iloc[0].value

    y_prev = y_start_neg
    for i, point in (
        w[(w["abs_rank"] < n_features) & (w["value"] < 0)].reset_index().iterrows()
    ):
        y_loc = y_prev + y_repel_coef if i != 0 else y_start_neg
        y_loc = max(point["value"], y_loc) if attract_to_points else y_loc
        ax.text(
            w.shape[0] - x_rank_offset_neg,
            y_loc,
            point["feature"],
            horizontalalignment="left",
            size=label_size,
            color="black",
            weight="regular",
        )
        y_prev = y_loc

    # Set plot axes labels
    factor_label = f"Factor{factor+1}" if isinstance(factor, int) else factor
    ax.set(ylabel=f"{factor_label} weight", xlabel="Feature rank")

    return ax


def plot_weights_scaled(
    model: mofa_model,
    x="Factor1",
    y="Factor2",
    view=0,
    n_features: int = 10,
    w_scaled: bool = True,
    label_size=5,
    y_repel_coef=0.05,
    attract_to_points=True,
    **kwargs,
):
    """
    Scatterplot of feature weights for two factors

    Parameters
    ----------
    model : mofa_model
        Factor model
    factor : optional
        Factor to use (default is Factor1)
    view : options
        The view to get the factors weights for (first view by default)
    n_features : optional
        Number of features to label with most positive and most negative weights
    label_size : optional
        Font size of feature labels (default is 5)
    y_repel_coef : optional
        Parameter to repel feature labels along the y axis (0.03 by default)
    attract_to_points : optional
        If place labels according to the Y coordinate of the point (False by default)
    """
    w = model.get_weights(views=view, factors=[x, y], df=True)
    w.columns = ["x", "y"]

    if w_scaled:
        w.x = w.x / abs(w.loc[abs(w.x).idxmax()].x)
        w.y = w.y / abs(w.loc[abs(w.y).idxmax()].y)

    wm = (
        w.rename_axis("feature")
        .reset_index()
        .melt(var_name="factor", id_vars=["feature"])
        .assign(
            value_abs=lambda x: np.abs(x.value), value_sign=lambda x: np.sign(x.value)
        )
        .sort_values("value_abs", ascending=False)
        .head(n_features)
        .sort_values(["factor", "value_sign"], ascending=True)
        .drop_duplicates("feature")
    )

    top_features = wm.sort_values("factor", ascending=True).feature.values

    # Construct the plot
    ax = sns.scatterplot("x", "y", data=w, linewidth=0, color="#CCCCCC", **kwargs)
    ax.set_xlim(-1.5, 1.5)
    ax.set_ylim(-1.5, 1.5)
    ax.set_aspect(1)
    for factor in wm.factor.unique():
        for sign in wm[wm.factor == factor].value_sign.unique():
            feature_set = wm[
                (wm.factor == factor) & (wm.value_sign == sign)
            ].feature.values
            w_set = w.loc[feature_set].sort_values("y", ascending=False)
            y_start_pos = w_set.y.max()
            y_prev = y_start_pos
            for i, row in enumerate(w_set.iterrows()):
                name, point = row
                y_loc = y_prev - y_repel_coef if i != 0 else y_start_pos
                y_loc = min(point.y, y_loc) if attract_to_points else y_loc
                y_prev = y_loc
                ax.text(point.x, y_loc, str(name), size=label_size)
                ax.plot([0, point.x], [0, point.y], linewidth=0.5, color="#333333")

    sns.despine(offset=10, trim=True, ax=ax)
    ax.set_xticks(np.arange(-1, 2.0, step=1.0))
    ax.set_yticks(np.arange(-1, 2.0, step=1.0))

    # Set plot axes labels
    x_factor_label = f"Factor{x+1}" if isinstance(x, int) else x
    y_factor_label = f"Factor{y+1}" if isinstance(y, int) else y
    ax.set(xlabel=f"{x_factor_label} weight", ylabel=f"{y_factor_label} weight")

    return ax


def plot_weights_heatmap(
    model: mofa_model,
    factors: Union[int, List[int]] = None,
    view=0,
    n_features: int = None,
    w_threshold: float = None,
    w_abs: bool = False,
    only_positive: bool = False,
    only_negative: bool = False,
    features_col: pd.DataFrame = None,
    cmap=None,
    xticklabels_size=10,
    yticklabels_size=None,
    cluster_factors=True,
    cluster_features=True,
    **kwargs,
):
    """
    Plot weights for top features in a heatmap

    Parameters
    ----------
    model : mofa_model
        Factor model
    factors : optional
        Factors to use (all factors in the model by default)
    view : options
        The view to get the factors weights for (first view by default)
    n_features : optional
        Number of features for each factor by their absolute value (10 by default)
    w_threshold : optional
        Absolute weight threshold for a feature to plot (no threshold by default)
    w_abs : optional
        If to plot absolute weight values
    only_positive : optional
        If to plot only positive weights
    only_negative : optional
        If to plot only negative weights
    features_col : optional
        Pandas data frame with index by feature name with the first column
        containing the colour for every feature
    cmap : optional
        Color map (blue-to-red divergent palette with by default)
    xticklabels_size : optional
        Font size for features labels (default is 10)
    yticklabels_size : optional
        Font size for factors labels (default is None)
    cluster_factors : optional
        If cluster factors (in rows; default is True)
    cluster_features : optional
        If cluster features (in columns; default in True)
    """

    # Set defaults
    n_features_default = 10
    if factors is None:
        factors = list(range(model.nfactors))
    if cmap is None:
        cmap = sns.diverging_palette(240, 10, n=9, as_cmap=True)

    # Fetch weights for the relevant factors
    w = (
        model.get_weights(views=view, factors=factors, df=True, absolute_values=w_abs)
        .rename_axis("feature")
        .reset_index()
    )
    wm = w.melt(id_vars="feature", var_name="factor", value_name="value")
    wm = wm.assign(value_abs=lambda x: x.value.abs())
    wm["factor"] = wm["factor"].astype("category")

    if only_positive and only_negative:
        print("Please specify either only_positive or only_negative")
        sys.exit(1)
    elif only_positive:
        wm = wm[wm.value > 0]
    elif only_negative:
        wm = wm[wm.value < 0]

    if n_features is None and w_threshold is not None:
        features = wm[wm.value_abs >= w_threshold].feature.unique()
    else:
        if n_features is None:
            n_features = n_features_default
        # Get a subset of features
        wm = wm.sort_values(["factor", "value_abs"], ascending=False).groupby("factor")
        if w_threshold is None:
            features = wm.head(n_features).feature.unique()
        else:
            features = wm[wm.value_abs >= w_threshold].head(n_features).feature.unique()

    w = w[w.feature.isin(features)].set_index("feature").T

    col_colors = features_col.loc[features, :] if features_col is not None else None

    if not isinstance(factors, Iterable) or len(factors) < 2:
        cluster_factors = False

    cg = sns.clustermap(
        w,
        cmap=cmap,
        col_colors=col_colors,
        xticklabels=True,
        row_cluster=cluster_factors,
        col_cluster=cluster_features,
        **kwargs,
    )

    cg.ax_heatmap.set_xticklabels(
        cg.ax_heatmap.xaxis.get_ticklabels(), rotation=90, size=xticklabels_size
    )
    cg.ax_heatmap.set_yticklabels(
        cg.ax_heatmap.yaxis.get_ticklabels(), rotation=0, size=yticklabels_size
    )

    return cg


def plot_weights_dotplot(
    model: mofa_model,
    factors: Union[int, List[int]] = None,
    view=0,
    n_features: int = None,
    w_threshold: float = None,
    w_abs: bool = False,
    only_positive: bool = False,
    only_negative: bool = False,
    palette=None,
    size: int = 30,
    linewidth: int = 1,
    xticklabels_size=8,
    yticklabels_size=5,
    ncols=1,
    sharex=True,
    sharey=False,
    **kwargs,
):
    """
    Plot weights for top features in a heatmap

    Parameters
    ----------
    model : mofa_model
        Factor model
    factors : optional
        Factors to use (all factors in the model by default)
    view : options
        The view to get the factors weights for (first view by default)
    n_features : optional
        Number of features for each factor by their absolute value (5 by default)
    w_threshold : optional
        Absolute weight threshold for a feature to plot (no threshold by default)
    w_abs : optional
        If to plot absolute weight values
    only_positive : optional
        If to plot only positive weights
    only_negative : optional
        If to plot only negative weights
    palette : optional
        Color map (blue-to-red divergent palette with by default)
    size : optional
        Dot size (default in 30)
    lienwidth : optional
        Dot outline width (default is 1)
    xticklabels_size : optional
        Font size for features labels (default is 10)
    yticklabels_size : optional
        Font size for factors labels (default is None)
    ncols : optional
        Number of columns when plotting multiple views (default is 1)
    sharex : bool
        If to use the same X axis across panels (True by default)
    sharey : bool
        If to use the same Y axis across panels (False by default)
    """

    # Set defaults
    n_features_default = 5
    if factors is None:
        factors = list(range(model.nfactors))
    if palette is None:
        palette = sns.diverging_palette(240, 10, n=9, as_cmap=True)

    # Fetch weights for the relevant factors
    w = (
        model.get_weights(views=view, factors=factors, df=True, absolute_values=w_abs)
        .rename_axis("feature")
        .join(model.features_metadata.loc[:, ["view"]])
        .reset_index()
    )
    wm = w.melt(id_vars=["feature", "view"], var_name="factor", value_name="value")
    wm = wm.assign(value_abs=lambda x: x.value.abs())
    wm["factor"] = wm["factor"].astype("category")

    if only_positive and only_negative:
        print("Please specify either only_positive or only_negative")
        sys.exit(1)
    elif only_positive:
        wm = wm[wm.value > 0]
    elif only_negative:
        wm = wm[wm.value < 0]

    # Fix factors order
    wm.factor = wm.factor.astype("category")
    wm.factor = wm.factor.cat.reorder_categories(
        sorted(wm.factor.cat.categories, key=lambda x: int(x.split("Factor")[1]))
    )
    wm.sort_values("factor")

    if n_features is None and w_threshold is not None:
        features = wm[wm.value_abs >= w_threshold].feature.unique()
    else:
        if n_features is None:
            n_features = n_features_default
        # Get a subset of features
        wm_g = wm.sort_values(["factor", "value_abs"], ascending=False).groupby(
            ["factor", "view"]
        )
        if w_threshold is None:
            features = wm_g.head(n_features).feature.unique()
        else:
            features = (
                wm_g[wm_g.value_abs >= w_threshold].head(n_features).feature.unique()
            )

    wm = wm[wm.feature.isin(features)]

    # Fix features order
    wm.feature = wm.feature.astype("category")
    wm.feature = wm.feature.cat.reorder_categories(features)

    wm = wm.sort_values(["factor", "feature"])

    # Figure out rows & columns for the grid with plots (one plot per view)
    view_vars = wm.view.unique()
    ncols = min(ncols, len(view_vars))
    nrows = int(np.ceil(len(view_vars) / ncols))
    fig, axes = plt.subplots(
        nrows,
        ncols,
        sharex=sharex,
        sharey=sharey,
        figsize=(
            ncols * rcParams["figure.figsize"][0],
            nrows * rcParams["figure.figsize"][1],
        ),
    )
    if ncols == 1:
        axes = np.array(axes).reshape(-1, 1)
    if nrows == 1:
        axes = np.array(axes).reshape(1, -1)

    for m, view in enumerate(view_vars):
        ri = m // ncols
        ci = m % ncols

        wm_view = wm.query("view == @view")

        # Construct the plot
        g = sns.scatterplot(
            data=wm_view,
            x="factor",
            y="feature",
            hue="value",
            linewidth=linewidth,
            s=size,
            palette=palette,
            ax=axes[ri, ci],
            **kwargs,
        )
        sns.despine(offset=10, trim=True, ax=g)
        g.legend().remove()

        norm = plt.Normalize(wm_view.value.min(), wm_view.value.max())
        cmap = (
            palette
            if palette is not None
            else sns.diverging_palette(220, 20, as_cmap=True)
        )
        sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
        sm.set_array([])
        try:
            g.figure.colorbar(sm, ax=axes[ri, ci])
            g.get_legend().remove()
        except Exception:
            warn("Cannot make a proper colorbar")

        plt.draw()

        g.set_title(view)

        g.set_xticklabels(g.get_xticklabels(), rotation=90, size=xticklabels_size)
        g.set_yticklabels(g.get_yticklabels(), size=yticklabels_size)

    # Remove unused axes
    for i in range(len(view_vars), ncols * nrows):
        ri = i // ncols
        ci = i % ncols
        fig.delaxes(axes[ri, ci])

    return g


def plot_weights_scatter(
    model: mofa_model,
    x="Factor1",
    y="Factor2",
    view=0,
    hist=False,
    n_features: int = 10,
    label_size: int = 5,
    **kwargs,
):
    """
    Plot weights for two factors

    Parameters
    ----------
    model : mofa_model
        Factor model
    x : optional
        Factor which weights to plot along X axis (Factor1 by default)
    y : optional
        Factor which weights to plot along Y axis (Factor2 by default)
    view : options
        The view to get the factors weights for (first view by default)
    hist : optional
        Boolean value if to add marginal histograms to the scatterplot (jointplot)
    n_features : optional
        Number of features to label (default is 10)
    label_size : optional
        Font size of feature labels (default is 5)
    """
    w = (
        model.get_weights(views=view, factors=[x, y], df=True)
        .rename_axis("feature")
        .reset_index()
    )

    # Get features to label
    wm = w.melt(id_vars="feature", var_name="factor", value_name="value")
    wm = wm.assign(value_abs=lambda x: x.value.abs())
    wm["factor"] = wm["factor"].astype("category")

    # Set default colour to darkgrey if none set
    if "c" not in kwargs and "color" not in kwargs:
        kwargs["color"] = "darkgrey"

    sns_plot = sns.jointplot if hist else sns.scatterplot
    plot = sns_plot(x=x, y=y, data=w, **kwargs)
    sns.despine(offset=10, trim=True)

    # Label some features
    add_text = plot.ax_joint.text if hist else plot.text
    if n_features is not None and n_features > 0:
        # Get a subset of features
        wm = wm.sort_values(["factor", "value_abs"], ascending=False).groupby("factor")
        features = wm.head(n_features).feature.unique()
        w_label = w[w.feature.isin(features)].set_index("feature")
        del wm

        # Add labels to the plot
        for i, point in w_label.iterrows():
            add_text(
                point[x],
                point[y],
                point.name,
                horizontalalignment="left",
                size=label_size,
                color="black",
                weight="regular",
            )

    return plot


def plot_weights_correlation(
    model: mofa_model,
    factors: Optional[Union[int, List[int]]] = None,
    views=None,
    covariates=None,
    linewidths=0,
    diag=False,
    full=True,
    cmap=None,
    square=True,
    **kwargs,
):
    """
    Plot correlation of weights and, if provided, covariates

    Parameters
    ----------
    model : mofa_model
        Factor model
    factors : optional
        Index of a factor (or indices of factors) to use (all factors by default)
    groups : optional
        Subset of groups to consider
    covarites : optional
        A vector, a matrix, or a data frame with covariates (one per column)
    linewidths : optional
        Heatmap linewidths argument (default is 0)
    diag : optional
        If to only plot lower triangle of the correlation matrix (False by default)
    full : optional
        If covariates are provided, also plot inter-factor and inter-covariates correlation coefficients (True by default)
    square : optional
        Heatmap square argument (True by default)
    cmap : optional
        Heatmap cmap argument
    """

    w = model.get_weights(factors=factors, views=views)
    if covariates is not None:
        # Transform a vector to a matrix
        if len(covariates.shape) == 1:
            covariates = pd.DataFrame(covariates)
        corr = np.corrcoef(w.T, covariates.T)
    else:
        corr = np.corrcoef(w.T)

    if covariates is not None:
        if not full:
            n_cov = covariates.shape[1]
            corr = corr[0:-n_cov, -n_cov:]

    mask = None
    if diag:
        # Generate a mask for the upper triangle
        mask = np.triu(np.ones_like(corr, dtype=np.bool))

    # Set up the matplotlib figure
    f, ax = plt.subplots(figsize=(11, 9))

    if cmap is None:
        # Generate a custom diverging colormap
        cmap = sns.diverging_palette(220, 10, as_cmap=True)

    # Generate labels for the heatmap
    if factors is None:
        factors = range(w.shape[1])
    fnames = [f"Factor{fi+1}" if isinstance(fi, int) else fi for fi in factors]
    if covariates is not None:
        if isinstance(covariates, pd.DataFrame):
            cnames = covariates.columns.values
        else:
            cnames = [f"Covar{ci+1}" for ci in covariates.shape[1]]
        xticklabels = cnames if not full else np.concatenate((fnames, cnames))
        yticklabels = fnames if not full else np.concatenate((fnames, cnames))
    else:
        xticklabels = fnames
        yticklabels = fnames

    # Draw the heatmap with the mask and correct aspect ratio
    g = sns.heatmap(
        corr,
        cmap=cmap,
        mask=mask,
        center=0,
        square=True,
        linewidths=0.5,
        xticklabels=xticklabels,
        yticklabels=yticklabels,
        cbar_kws={"shrink": 0.5},
        **kwargs,
    )

    g.set_yticklabels(g.yaxis.get_ticklabels(), rotation=0)

    return g



from .core import mofa_model
from .utils import *

import sys
from warnings import warn
from typing import Union, Optional, List, Iterable, Sequence
from functools import partial

import numpy as np
from scipy.stats import pearsonr
import pandas as pd
from pandas.api.types import is_numeric_dtype
import matplotlib.pyplot as plt
from matplotlib import rcParams
import seaborn as sns

from .utils import maybe_factor_indices_to_factors, _make_iterable, _is_iter
from .plot_utils import _plot_grid


### VARIANCE EXPLAINED ###


def plot_r2(
    model: mofa_model,
    x="Group",
    y="Factor",
    factors: Union[int, List[int], str, List[str]] = None,
    groups_df: pd.DataFrame = None,
    group_label: str = None,
    views=None,
    groups=None,
    cmap="Blues",
    vmin=None,
    vmax=None,
    **kwargs,
):
    """
    Plot R2 values for the model

    Parameters
    ----------
    model : mofa_model
        Factor model
    x : str
        Dimension along X axis: Group (default), View, or Factor
    y : str
        Dimension along Y axis: Group, View, or Factor (default)
    factors : optional
        Index of a factor (or indices of factors) to use (all factors by default)
    views : optional
        Make a plot for certain views (None by default to plot all views)
    groups : optional
        Make a plot for certain groups (None by default to plot all groups)
    group_label : optional
        Sample (cell) metadata column to be used as group assignment
    groups_df : optional pd.DataFrame
        Data frame with samples (cells) as index and first column as group assignment
    cmap : optional
        The colourmap for the heatmap (default is 'Blues' with darker colour for higher R2)
    vmin : optional
        Display all R2 values smaller than vmin as vmin (0 by default)
    vmax : optional
        Display all R2 values larger than vmax as vmax (derived from the data by default)
    """
    r2 = model.get_r2(
        factors=factors,
        groups=groups,
        views=views,
        group_label=group_label,
        groups_df=groups_df,
    )

    vmax = r2.R2.max() if vmax is None else vmax
    vmin = 0 if vmin is None else vmin

    split_by = [dim for dim in ["Group", "View", "Factor"] if dim not in [x, y]]
    assert (
        len(split_by) == 1
    ), "x and y values should be different and be one of Group, View, or Factor"
    split_by = split_by[0]

    split_by_items = r2[split_by].unique()
    fig, axes = plt.subplots(ncols=len(split_by_items), sharex=True, sharey=True)
    cbar_ax = fig.add_axes([0.91, 0.3, 0.03, 0.4])
    if len(split_by_items) == 1:
        axes = [axes]

    for i, item in enumerate(split_by_items):
        r2_sub = r2[r2[split_by] == item]
        r2_df = r2_sub.sort_values("R2").pivot(index=y, columns=x, values="R2")

        if y == "Factor":
            # Sort by factor index
            r2_df.index = r2_df.index.astype("category")
            r2_df.index = r2_df.index.reorder_categories(
                sorted(r2_df.index.categories, key=lambda x: int(x.split("Factor")[1]))
            )
            r2_df = r2_df.sort_values("Factor")

        if x == "Factor":
            # Re-order columns by factor index
            r2_df.columns = r2_df.columns.astype("category")
            r2_df.columns = r2_df.columns.reorder_categories(
                sorted(
                    r2_df.columns.categories, key=lambda x: int(x.split("Factor")[1])
                )
            )
            r2_df = r2_df[r2_df.columns.sort_values()]

        g = sns.heatmap(
            r2_df.sort_index(level=0, ascending=False),
            ax=axes[i],
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            cbar=i == 0,
            cbar_ax=None if i else cbar_ax,
            **kwargs,
        )

        axes[i].set_title(item)
        axes[i].tick_params(axis="both", which="both", length=0)
        if i == 0:
            g.set_yticklabels(g.yaxis.get_ticklabels(), rotation=0)
        else:
            axes[i].set_ylabel("")

    plt.close()
    return fig


def plot_r2_pvalues(
    model: mofa_model,
    factors: Union[int, List[int], str, List[str]] = None,
    n_iter: int = 100,
    groups_df: pd.DataFrame = None,
    group_label: str = None,
    view=0,
    fdr: bool = True,
    cmap="binary_r",
    **kwargs,
):
    """
    Plot R2 values for the model

    Parameters
    ----------
    model : mofa_model
        Factor model
    factors : optional
        Index of a factor (or indices of factors) to use (all factors by default)
    view : optional
        Make a plot for a cetrain view (first view by default)
    groups_df : optional pd.DataFrame
        Data frame with samples (cells) as index and first column as group assignment
    group_label : optional
        Sample (cell) metadata column to be used as group assignment
    fdr : optional bool
        If plot corrected PValues (FDR)
    cmap : optional
        The colourmap for the heatmap (default is 'binary_r' with darker colour for smaller PValues)
    """
    r2 = model.get_r2_null(
        factors=factors,
        groups_df=groups_df,
        group_label=group_label,
        n_iter=n_iter,
        return_pvalues=True,
        fdr=fdr,
    )
    pvalue_column = "FDR" if fdr else "PValue"
    # Select a certain view if necessary
    if view is not None:
        view = model.views[view] if isinstance(view, int) else view
        r2 = r2[r2["View"] == view]
    r2_df = r2.sort_values("PValue").pivot(
        index="Factor", columns="Group", values=pvalue_column
    )

    # Sort by factor index
    r2_df.index = r2_df.index.astype("category")
    r2_df.index = r2_df.index.reorder_categories(
        sorted(r2_df.index.categories, key=lambda x: int(x.split("Factor")[1]))
    )
    r2_df = r2_df.sort_values("Factor")

    g = sns.heatmap(r2_df.sort_index(level=0, ascending=False), cmap=cmap, **kwargs)

    g.set_yticklabels(g.yaxis.get_ticklabels(), rotation=0)

    return g


def plot_r2_barplot(
    model: mofa_model,
    factors: Union[int, List[int], str, List[str]] = None,
    view=0,
    groups_df: pd.DataFrame = None,
    group_label: str = None,
    x="Factor",
    y="R2",
    groupby="Group",
    xticklabels_size=10,
    linewidth=0,
    stacked=False,
    **kwargs,
):
    """
    Plot R2 values for the model

    Parameters
    ----------
    model : mofa_model
        Factor model
    factors : optional
        Index of a factor (or indices of factors) to use (all factors by default)
    view : optional
        Make a plot for a cetrain view (first view by default)
    groups_df : optional pd.DataFrame
        Data frame with samples (cells) as index and first column as group assignment
    group_label : optional
        Sample (cell) metadata column to be used as group assignment
    x : optional
        Value to plot along the x axis (default is Factor)
    y : optional
        Value to plot along the y axis (default is R2)
    groupby : optional
        Column to group bars for R2 values by (default is Group)
    xticklabels_size : optional
        Font size for group labels (default is 10)
    linewidth : optional
        Linewidth (0 by default)
    stacked : optional
        Plot a stacked barplot instead of a grouped barplot
    """
    r2 = model.get_r2(factors=factors, groups_df=groups_df, group_label=group_label)
    # Select a certain view if necessary
    if view is not None:
        view = model.views[view] if isinstance(view, int) else view
        r2 = r2[r2["View"] == view]
    r2_df = r2.sort_values("R2")

    # Sort by factor index
    r2_df.Factor = r2_df.Factor.astype("category")
    r2_df.Factor = r2_df.Factor.cat.reorder_categories(
        sorted(r2_df.Factor.cat.categories, key=lambda x: int(x.split("Factor")[1]))
    )
    r2_df = r2_df.sort_values("Factor")

    if stacked:
        g = r2_df.pivot(index="Factor", columns="Group", values="R2").plot(
            kind="bar", stacked=True, linewidth=linewidth, **kwargs
        )
        plt.ylabel("R2")
        plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.0)
    else:
        # Grouped barplot
        g = sns.barplot(
            data=r2_df.sort_index(level=0, ascending=False),
            x=x,
            y=y,
            hue=groupby,
            linewidth=linewidth,
            **kwargs,
        )

        g.set_xticklabels(g.xaxis.get_ticklabels(), rotation=90, size=xticklabels_size)

    return g



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mofax']

package_data = \
{'': ['*']}

install_requires = \
['h5py>=2.9.0',
 'matplotlib>=3.1.0',
 'numpy>=1.17.0',
 'pandas>=1.0.0',
 'seaborn>=0.9.0']

setup_kwargs = {
    'name': 'mofax',
    'version': '0.3.6',
    'description': 'Work with MOFA+ models in Python',
    'long_description': None,
    'author': 'Danila Bredikhin',
    'author_email': 'danila.bredikhin@embl.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)
